/*
Fluffy Tails
	by Lach

beta
*/

local SKIN = "fluffy" // skin name to run this script for

local WALK_SPEED = 28*FRACUNIT // normalspeed before hitting "dashmode"
local WALK_TIME = TICRATE // time to walk above runspeed before "dashmode" will activate

local FLIGHTCHARGE_THRESHOLD = 3*TICRATE/2 // time to fully charge a flight charge

local FLIGHT_TIME = 3*TICRATE // total flight time
local FLIGHT_ACCELERATION = FRACUNIT // acceleration while flying
local FLIGHT_DECELERATION = FRACUNIT // deceleration when flying upwards
local FLIGHT_THRUST = FRACUNIT/3 // upwards acceleration while vertical momentum is positive
local FLIGHT_THRUST_BOOST = 2*FRACUNIT/3 // upwards acceleration while vertical momentum is negative (recovering from falling)
local FLIGHT_SPIN_THRUST = FRACUNIT/3 // landing aid, downwards thrust when spin is held during flight
local FLIGHT_DECELERATION_THRESHOLD = FRACUNIT // vertical speed at which horizontal deceleration starts

local FLICKY_ACCEL_H = FRACUNIT // flicky horizontal acceleration
local FLICKY_ACCEL_Z = FRACUNIT/2 // flicky vertical acceleration
local FLICKY_ATTACK_RADIUS = 768*FRACUNIT // distance from Fluffy within which enemies will be attacked by Flickies
local FLICKY_ATTACK_INTERVAL = 2*TICRATE // a flicky will attempt an attack every time this amount of time has passed
local FLICKY_ATTACK_FUSE = 3*TICRATE // a flicky will give up on its attack this amount of time after starting
local FLICKY_COLORS = {SKINCOLOR_BLUE, SKINCOLOR_TAFFY, SKINCOLOR_CARDINAL, SKINCOLOR_BEEG}
local FLICKY_SPRITES = {SPR_3DFB, SPR_3DFP, SPR_3DFR, SPR_3DFG}
local NUMFLICKIES = #FLICKY_SPRITES

rawset(_G, "CC_FLICKY_SPRITES", FLICKY_SPRITES) // needed in CHRISPYCHARS.lua
rawset(_G, "CC_FLICKY_COLORS", FLICKY_COLORS) // ditto

local FM_WALK = 1
local FM_DASH = 2

local ANIM_NONE = 1
local ANIM_IDLE = 2
local ANIM_TWIRL = 3
local ANIM_RUN = 4
local ANIM_FLY = 5
local ANIM_DAB = 6
local ANIM_STOP = 7
local ANIM_HANG = 8
local ANIM_TPOSE = 9

local TAIL_ANIMS = { // table of tail animations
	[SPR_PLAY] = {
		[SPR2_STND] = ANIM_IDLE,
		[SPR2_WAIT] = ANIM_IDLE,
		[SPR2_LAND] = ANIM_IDLE,
		[SPR2_BODY] = ANIM_IDLE,
		[SPR2_EDGE] = ANIM_IDLE,
		[SPR2_DNCE] = ANIM_IDLE,
		[SPR2_CNT1] = ANIM_IDLE,
		[SPR2_TRNS] = ANIM_IDLE,
		[SPR2_FAN_] = ANIM_TWIRL,
		[SPR2_SPIN] = ANIM_TWIRL,
		[SPR2_WALK] = ANIM_TWIRL,
		[SPR2_ROLL] = ANIM_TWIRL,
		[SPR2_TWIN] = ANIM_TWIRL,
		[SPR2_SPNG] = ANIM_TWIRL,
		[SPR2_FALL] = ANIM_TWIRL,
		[SPR2_SWIM] = ANIM_TWIRL,
		[SPR2_FLT] = ANIM_TWIRL,
		[SPR2_RUN] = ANIM_RUN,
		[SPR2_SKID] = ANIM_STOP,
		[SPR2_GASP] = ANIM_STOP,
		[SPR2_FLY] = ANIM_FLY,
		[SPR2_CARY] = ANIM_FLY,
		[SPR2_TIRE] = ANIM_FLY,
		[SPR2_POSE] = ANIM_DAB,
		[SPR2_DEAD] = ANIM_TWIRL,
		[SPR2_DRWN] = ANIM_TWIRL,
		[SPR2_PDED] = ANIM_TWIRL,
		[SPR2_PAIN] = ANIM_HANG,
		[SPR2_RIDE] = ANIM_HANG,
		[SPR2_TPOS] = ANIM_TPOSE,
		
		// extra poses, I dunno how well they'll work
		[SPR2_CLMB] = ANIM_TWIRL,
		[SPR2_CLNG] = ANIM_TWIRL,
		[SPR2_DASH] = ANIM_RUN,
		[SPR2_GLID] = ANIM_RUN,
		[SPR2_BATL] = ANIM_NONE
	}
}

local TAIL_STATES = { // table of states to use depending on the animation
	[ANIM_NONE] = S_INVISIBLE, // default; empty sprite
	[ANIM_IDLE] = S_FLUFFYTAIL_IDLE,
	[ANIM_TWIRL] = S_FLUFFYTAIL_TWIRL,
	[ANIM_RUN] = S_FLUFFYTAIL_RUN,
	[ANIM_FLY] = S_FLUFFYTAIL_FLY,
	[ANIM_DAB] = S_FLUFFYTAIL_DAB,
	[ANIM_STOP] = S_FLUFFYTAIL_STOP,
	[ANIM_HANG] = S_FLUFFYTAIL_TWIRL_BELOW,
	[ANIM_TPOSE] = S_FLUFFYTAIL_TPOSE
}

if not CC_NeckSkins
	rawset(_G, "CC_NeckSkins", {})
end
CC_NeckSkins[SKIN] = { // Neck support
	skin = SKIN,
	speed = FRACUNIT/5,
	pivot = true,
	pitches = 5,
	idleangle = -ANGLE_45,
	idlepitch = ANG30,
	emoteoffset = 6*FRACUNIT,
	eastereggname = "\102\x6c\117\x66\102\x20\101\x2d\116\x34\049\x6c\053",
	trackables = {},
}
for _, item in ipairs(CC_NeckSkins["pointy"].trackables) // Use all of Pointy's trackables except for sneezes
	if not item.sneeze
		table.insert(CC_NeckSkins[SKIN].trackables, item)
	end
end

if not CC_TrickSkins 
	rawset(_G, "CC_TrickSkins", {})
end
CC_TrickSkins[SKIN] = { // Tricks support
	bouncetricks = true,
	{
		sprite2 = SPR2_SPNG,
		tics = 0,
		rotation = 1,
	},
	{
		sprite2 = SPR2_GASP,
		tics = 0,
		rotation = 0,
	},
	{
		rarity = 2*FRACUNIT,
		sprite2 = SPR2_SWIM,
		tics = 2,
		rotation = "random",
		frames = 4,
	},
	{
		sprite2 = SPR2_EDGE,
		tics = 4,
		rotation = 0,
		frames = 2,
	},
	{
		sprite2 = SPR2_FAN_,
		rotation = "random",
		tics = 0,
	},
	{
		sprite2 = SPR2_WAIT,
		rotation = 0,
		angle = ANGLE_180,
		tics = 4,
		frames = 2,
	},
	{
		sprite2 = SPR2_RIDE,
		rotation = 0,
		tics = 0,
	},
	{
		sprite2 = SPR2_WALK,
		rotation = 0,
		tics = 2,
		frames = 8,
	}
}

if not CC_DanceSkins
	rawset(_G, "CC_DanceSkins", {})
end
CC_DanceSkins[SKIN] = {
	[1] = { // Floss 1
		condition = function(player)
			local chrispy = player.chrispy
			return not (chrispy and chrispy.flags & (CF_BOUNCING|CF_CROUCHING))
		end,
		keys = {
			{
				sprite2 = SPR2_DNCE,
				frame = A,
				tics = 5,
				angle = ANGLE_180
			},
			{
				sprite2 = SPR2_DNCE,
				frame = B,
				tics = 5,
				angle = ANGLE_180,
				queuetransition = true
			},
			{
				sprite2 = SPR2_DNCE,
				frame = C,
				tics = 5,
				angle = ANGLE_180
			},
			{
				sprite2 = SPR2_DNCE,
				frame = D,
				tics = 5,
				angle = ANGLE_180,
				queuetransition = true
			}
		}
	},
	[2] = { // Floss 2
		condition = function(player)
			local chrispy = player.chrispy
			return not (chrispy and chrispy.flags & (CF_BOUNCING|CF_CROUCHING))
		end,
		keys = {
			{
				sprite2 = SPR2_DNCE,
				frame = E,
				tics = 5,
				angle = ANGLE_180
			},
			{
				sprite2 = SPR2_DNCE,
				frame = F,
				tics = 5,
				angle = ANGLE_180,
				queuetransition = true
			},
			{
				sprite2 = SPR2_DNCE,
				frame = G,
				tics = 5,
				angle = ANGLE_180
			},
			{
				sprite2 = SPR2_DNCE,
				frame = H,
				tics = 5,
				angle = ANGLE_180,
				queuetransition = true
			}
		}
	}
}

CC_Footsteps[SKIN] = {
	[SPR2_BODY] = {
		flags = FSF_NOTRANSITION,
		[A] = sfx_rstep1,
	},
	[SPR2_WALK] = {
		[C] = sfx_rstep1,
		[G] = sfx_rstep2,
	},
	[SPR2_ROLL] = {
		flags = FSF_MOVEMENTONLY,
		repeatsound = sfx_marble,
	},
	[SPR2_SPIN] = {
		flags = FSF_MOVEMENTONLY,
		repeatsound = sfx_marble,
	},
	[SPR2_POSE] = {
		flags = FSF_NOTRANSITION|FSF_MOVEMENTONLY,
		[A] = sfx_rstep1*FRACUNIT + sfx_rstep2,
		repeatsound = sfx_squeal,
	},
	[SPR2_RUN] = {
		flags = FSF_AERIAL,
		repeatsound = 10*FRACUNIT + sfx_rput,
	},
	[SPR2_FLY] = {
		flags = FSF_AERIAL,
		repeatsound = 10*FRACUNIT + sfx_rput,
	},
	[SPR2_CARY] = {
		flags = FSF_AERIAL,
		repeatsound = 10*FRACUNIT + sfx_rput,
	},
	[SPR2_CNT1] = {
		flags = FSF_MOVEMENTONLY,
		repeatsound = sfx_rskid2
	}
}

local FLAGS = {
	CC_RegisterFlag("CF_FLYING"),
	CC_RegisterFlag("CF_TOSSING"),
	CC_RegisterFlag("CF_FLIGHTCHARGE"),
	CC_RegisterFlag("CF_FLICKYTARGET")
}
rawset(_G, "CF_FLUFFYMASK", FLAGS[#FLAGS]|(FLAGS[#FLAGS] - 1) - (FLAGS[1] - 1))
FLAGS = nil

local function valid(mo)
	return mo and mo.valid
end

local function KillFlightCharge(fluffy)
	fluffy.flightcharge = 0
	fluffy.player.charability2 = skins[fluffy.mo.skin].ability2
end

local flickytargets
local function FlickySearch(mo, item)
	if item.valid
	and item.health
	and item.flags & (MF_ENEMY|MF_BOSS)
	and item.flags & MF_SHOOTABLE
	and not (item.flags2 & MF2_FRET)
		table.insert(flickytargets, item)
	end
end

// This is literally just to support the script from Metallic Madness :V we wanted all the Summer 2020 OLDC levels playable
addHook("PreThinkFrame", do
	for player in players.iterate
		if not player.normalspeed_last
		or not player.mo
			continue
		end
		
		if player.powers[pw_tailsfly]
		and player.cmd.buttons & BT_JUMP
		and player.mo.skin == SKIN
			player.cmd.buttons = $ | BT_USE
		end
	end
end)

// Fuck youuuuuu SRB2
// This is so bot players don't copy Fluffy's thrustfactor of 0 during flight
// (it takes advantage of how player thinkers are executed in order -
// the player will use the stats from the previous frame's ThinkFrame and then PlayerThink will be called,
// setting them to these values, which the bot will copy. Then the values are set back to what the player should use in ThinkFrame)
addHook("PlayerThink", function(player)
	local mo = player.mo
	if player == players[0]
	and valid(players[1]) and players[1].bot
	and player.powers[pw_tailsfly]
	and mo and mo.skin == SKIN
		player.thrustfactor = players[1].thrustfactor
		player.normalspeed = players[1].normalspeed
	end
end)

// Main thinker

local function FluffyThinker(player)
	local mo = player.mo
	if not mo
	or not player.chrispy
		return
	end
	
	local chrispy = player.chrispy
	local fluffy = chrispy.fluffy
	
	if mo.skin ~= SKIN
		if fluffy
			chrispy.fluffy = nil
			chrispy.flags = $ & ~CF_FLUFFYMASK
		end
		return
	end
	
	if player.powers[pw_carry] == CR_NIGHTSMODE
		return
	end
	
	// okay, so this player is definitely playing as fluffy
	
	if not chrispy.fluffy // omg you just spawned a fluffy?! you lucky player let's set up your variables for you
		fluffy = {} // table used to store all fluffy-specific variables
		chrispy.fluffy = fluffy
		fluffy.player = player
		fluffy.mo = mo
		fluffy.roll = 0
		fluffy.lift = 0
		fluffy.mode = FM_DASH
		fluffy.walktime = 0
		fluffy.flightcharge = 0
		fluffy.dancepattern = 0
		fluffy.flickies = {}
		fluffy.flickydelay = 0
		chrispy.revheight = -13
		chrispy.superanims = false
		
		player.thrustfactor = skins[SKIN].thrustfactor
		player.charability2 = skins[SKIN].ability2
	end
	
	// locals
	
	local tails = mo.fluffytails
	local flip = P_MobjFlip(mo)
	local grounded = P_IsObjectOnGround(mo)
	local slope = mo.standingslope
	local looking = mo.state == S_PLAY_LOOK
	
	// flight charge
	
	if srb2p
		// doesn't work lol
	elseif mo.state == S_CHRISPY_CROUCH
	and not (chrispy.flags & CF_EMOTEWHEEL)
	and grounded
		if chrispy.buttons[BT_JUMP] == 1
		and not fluffy.flightcharge
			S_StartSound(mo, sfx_ftdash)
			fluffy.flightcharge = 1
			player.charability2 = CA2_NONE
		end
		
		if fluffy.flightcharge
			if fluffy.flightcharge > FLIGHTCHARGE_THRESHOLD
				KillFlightCharge(fluffy)
			elseif not chrispy.buttons[BT_JUMP]
				fluffy.flightcharge = $ - 1
				if not fluffy.flightcharge
					KillFlightCharge(fluffy)
				end
			else
				if fluffy.flightcharge == FLIGHTCHARGE_THRESHOLD
					mo.state = S_PLAY_STND
				else
					player.pflags = $ | PF_STASIS
				end
				fluffy.flightcharge = $ + 1
			end
		end
	elseif fluffy.flightcharge
		if fluffy.flightcharge <= FLIGHTCHARGE_THRESHOLD
		or mo.state == S_CHRISPY_DAB
		or player.powers[pw_carry]
		or P_PlayerInPain(player)
			KillFlightCharge(fluffy)
			player.pflags = $ | PF_JUMPDOWN
		elseif not chrispy.buttons[BT_JUMP]
			KillFlightCharge(fluffy)
			player.powers[pw_tailsfly] = tailsflytics
			player.pflags = $ & ~(PF_JUMPED|PF_NOJUMPDAMAGE|PF_SPINNING|PF_STARTDASH) | PF_THOKKED | PF_CANCARRY
			P_SetObjectMomZ(mo, 24*FRACUNIT)
			mo.z = $ + flip
			mo.state = S_PLAY_FLY
			S_StartSound(mo, sfx_zoom)
			CC_SpawnDustRing(mo, 6*mo.scale, not grounded and 2*mo.scale or 0)
		else
			player.pflags = $ | PF_JUMPSTASIS
		end
	end
	
	// "dashmode"
	
	local speed = FixedDiv(FixedHypot(mo.momx, mo.momy), mo.scale)
	if fluffy.mode == FM_WALK
	and player.runspeed ~= skins[SKIN].runspeed
		if speed >= FixedMul(skins[SKIN].runspeed, mo.movefactor)
		and mo.state == S_PLAY_WALK
		and grounded
		and not fluffy.flightcharge
		and not player.powers[pw_super]
			if fluffy.walktime < WALK_TIME
				if fluffy.walktime == WALK_TIME/2
					S_StartSound(mo, sfx_ftdash)
				end
				fluffy.walktime = $ + 1
			end
		else
			if fluffy.walktime > 0
				fluffy.walktime = $ - 1
			end
		end
		
		if fluffy.walktime >= WALK_TIME
		or speed > player.normalspeed + FRACUNIT
			player.normalspeed = skins[SKIN].normalspeed
			player.runspeed = skins[SKIN].runspeed
			fluffy.mode = FM_DASH
			fluffy.walktime = WALK_TIME
			if not player.powers[pw_super]
				if grounded
					for i = 1, 6
						SpawnChrispyDust(player, mo.scale, mo.scale >> 2, 0)
					end
				elseif not player.bot
					P_SpawnThokMobj(player)
				end
			end
		end
	else
		if speed < FixedMul(player.runspeed, mo.movefactor)
			fluffy.walktime = $ - 1
		else
			fluffy.walktime = WALK_TIME
		end
		
		if fluffy.walktime <= 0
			player.normalspeed = WALK_SPEED
			player.runspeed = 1024*FRACUNIT
			fluffy.mode = FM_WALK
		end
	end
	
	// Flight
	
	local roll = 0
	
	if player.panim == PA_ABILITY
	and player.charability == CA_FLY
		local in2d = twodlevel or mo.flags2 & MF2_TWOD or mo.currentaxis
		local flight = player.powers[pw_tailsfly]
		local topspeed = FixedMul(mo.eflags & MFE_UNDERWATER and skins[SKIN].normalspeed >> 1 or WALK_SPEED, mo.scale)
		local acceleration = FixedMul(FLIGHT_ACCELERATION, mo.scale)
		local deceleration = FixedMul(FLIGHT_DECELERATION, mo.scale)
		local oldmagnitude = FixedHypot(mo.momx - player.cmomx, mo.momy - player.cmomy)
		local newmagnitude
		local gravity = P_GetMobjGravity(mo)
		
		if not (chrispy.flags & CF_FLYING)
			chrispy.flags = $ | CF_FLYING
			player.sloperollangle_override = true
			flight = FLIGHT_TIME
		end
		
		if fluffy.lift
			fluffy.lift = $ - 1
		end
		
		player.normalspeed = 2048*FRACUNIT // LOLL I hope I've accounted for resetting this properly
		player.thrustfactor = 0
		player.fly1 = 0 // why is fly1 positive with SF_MULTIABILITY's flight? lol
		mo.momz = $ - gravity // zero gravity? :pog:
		
		// vertical movement
		if flight and (chrispy.buttons[BT_JUMP] or fluffy.lift)
			local momz = flip*mo.momz
			local power = 0
			
			// determine upwards thrust
			if momz <= 0
				power = FLIGHT_THRUST_BOOST
			elseif momz < topspeed >> 1
				power = FLIGHT_THRUST
				if mo.eflags & MFE_UNDERWATER
					power = $ >> 1
				end
			end
			P_SetObjectMomZ(mo, power, true)
			
			// decelerate horizontally if flying up fast enough
			if momz > FixedMul(FLIGHT_DECELERATION_THRESHOLD, mo.scale)
				topspeed = max($ - 2*max(momz, 0), 0)
				if oldmagnitude > topspeed
					newmagnitude = oldmagnitude - deceleration
					mo.momx = FixedMul(FixedDiv($ - player.cmomx, oldmagnitude), newmagnitude)
					mo.momy = FixedMul(FixedDiv($ - player.cmomy, oldmagnitude), newmagnitude)
					oldmagnitude = newmagnitude
				end
			end
			
			if not fluffy.lift
				fluffy.lift = 2
			end
		else
			mo.momz = $ + 3*gravity
			if flight
				if chrispy.buttons[BT_USE]
				and player.bot ~= 1
					P_SetObjectMomZ(mo, -FLIGHT_SPIN_THRUST / (mo.eflags & MFE_UNDERWATER and 2 or 1), true)
				else
					flight = $ + 1
				end
			end
		end
		
		// horizontal movement
		if (player.cmd.forwardmove or player.cmd.sidemove)
			local thrustdir
			if in2d
				thrustdir = mo.angle
			else
				thrustdir = (player.cmd.angleturn << 16) + R_PointToAngle2(0, 0, player.cmd.forwardmove << 16, -player.cmd.sidemove << 16)
			end
			
			if player.powers[pw_super] or player.powers[pw_sneakers]
				acceleration = 3 * $ / 2
				topspeed = $ << 1
			end
			P_Thrust(mo, thrustdir, acceleration)
			newmagnitude = FixedHypot(mo.momx - player.cmomx, mo.momy - player.cmomy)
			
			// horizontal speed cap
			if newmagnitude > topspeed // are we going faster than the cap?
				if oldmagnitude > topspeed // if we were going faster than the cap previously,
					if newmagnitude > oldmagnitude // did we gain speed?
						// if so, reduce the speed back to the old speed. This lets us change direction without accelerating.
						mo.momx = FixedMul(FixedDiv($ - player.cmomx, newmagnitude), oldmagnitude) + player.cmomx
						mo.momy = FixedMul(FixedDiv($ - player.cmomy, newmagnitude), oldmagnitude) + player.cmomy
					end
				else // if we only just now went over the speed cap,
					// reduce our speed to the top speed
					mo.momx = FixedMul(FixedDiv($ - player.cmomx, newmagnitude), topspeed) + player.cmomx
					mo.momy = FixedMul(FixedDiv($ - player.cmomy, newmagnitude), topspeed) + player.cmomy
				end
			end
		elseif player.pflags & PF_APPLYAUTOBRAKE // autobrake
			CC_Autobrake(player)
		end
		
		// swim duration
		if mo.state == S_PLAY_SWIM
			local d = P_GetPlayerControlDirection(player)
			if mo.tics > 4 - d
				mo.tics = 4 - d
			end
		end
		
		// you wanna fight i wanna tossle
		if chrispy.flags & CF_TOSSING
			local target = fluffy.target
			local cargo = fluffy.cargo
			local cargoplayer = valid(cargo) and cargo.player
			local untargeted = cargo == target
			
			if not valid(target) // no enemy? no cargo player? cancel throwing
			or not fluffy.target.health
			or not cargoplayer
			or cargo.tracer ~= mo
				chrispy.flags = $ & ~CF_TOSSING
				mo.state = S_PLAY_FLY
			elseif mo.state == S_FLUFFY_TOSS1 // turn towards the target
				if player.pflags & PF_DIRECTIONCHAR
					local angle
					if untargeted
						angle = cargo.angle
					else
						angle = R_PointToAngle2(cargo.x, cargo.y, target.x, target.y)
					end
					player.drawangle = fluffy.tossangle + (angle - fluffy.tossangle)/4
					fluffy.tossangle = player.drawangle
				end
			elseif mo.state == S_FLUFFY_TOSS2 // toss the player!
				local angle
				local hdist
				local zdist
				
				if untargeted
					angle = cargo.angle
					hdist = 32*FRACUNIT
					zdist = P_MobjFlip(cargo)*4*FRACUNIT
				else
					angle = R_PointToAngle2(cargo.x, cargo.y, target.x, target.y)
					hdist = FixedHypot(target.x - cargo.x, target.y - cargo.y)
					zdist = (target.z + (target.height >> 1)) - (cargo.z + (cargo.height >> 1))
				end
				
				local tossspeed = FixedMul(3*skins[mo.skin].normalspeed/5, mo.scale)
				local tossangle = R_PointToAngle2(0, 0, hdist, zdist)
				local hthrust = FixedMul(tossspeed, cos(tossangle))
				local airtime = hdist/hthrust
				local zthrust = zdist/airtime - airtime*P_GetMobjGravity(mo)
				
				cargoplayer.pflags = $ & ~(PF_NOJUMPDAMAGE|PF_THOKKED) | PF_JUMPED | PF_SPINNING
				cargoplayer.powers[pw_carry] = CR_NONE
				cargo.state = S_PLAY_ROLL
				cargo.tracer = nil
				P_InstaThrust(cargo, angle, hthrust)
				cargo.momz = zthrust
				P_Thrust(mo, angle, -hthrust/2)
				mo.momz = $ - zthrust/2
				S_StartSound(cargo, sfx_zoom)
				
				fluffy.tossangle = angle
				chrispy.flags = $ & ~CF_TOSSING
				player.pflags = $ & ~PF_CANCARRY
			else // idk something happened
				chrispy.flags = $ & ~CF_TOSSING
				mo.state = S_PLAY_FLY
			end
		end
		
		if mo.state == S_FLUFFY_TOSS2
		and player.pflags & PF_DIRECTIONCHAR
			player.drawangle = fluffy.tossangle
			if mo.tics == 1
				player.pflags = $ | PF_CANCARRY
			end
		end
		
		// gauge
		local gauge = fluffy.gauge
		
		if not valid(gauge)
			gauge = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_OVERLAY)
			gauge.target = mo
			gauge.state = S_FLUFFY_GAUGENEEDLE
			fluffy.gauge = gauge
			
			gauge = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_OVERLAY)
			gauge.target = fluffy.gauge
			gauge.state = S_FLUFFY_GAUGE
			fluffy.gauge.tracer = gauge
			
			gauge = fluffy.gauge
		end
		
		gauge.rollangle = (FLIGHT_TIME - flight) * (INT32_MAX/FLIGHT_TIME)
		
		gauge.flags2 = $ | MF2_DONTDRAW
		for _, p in pairs({displayplayer, secondarydisplayplayer})
			if (player.bot == 2
			or p.mo == mo
			or p.powers[pw_carry] == CR_PLAYER
			and p.mo.tracer == mo)
			and player.cc_gaugeenabled ~= false
			and not p.powers[pw_super]
				gauge.flags2 = $ & ~MF2_DONTDRAW
				break
			end
		end
		gauge.tracer.flags2 = gauge.flags2
		
		// yay we're done
		if flight
			roll = player.cmd.sidemove
		end
		if player.powers[pw_super]
			player.powers[pw_tailsfly] = FLIGHT_TIME
		else
			player.powers[pw_tailsfly] = flight
		end
	elseif chrispy.flags & CF_FLYING
		chrispy.flags = $ & ~(CF_FLYING|CF_TOSSING)
		player.sloperollangle_override = false
		player.thrustfactor = skins[SKIN].thrustfactor
		player.normalspeed = fluffy.mode == FM_WALK and WALK_SPEED or skins[SKIN].normalspeed
		
		if valid(fluffy.gauge)
			P_RemoveMobj(fluffy.gauge)
		end
	end
	
	if (fluffy.roll or roll)
	and not splitscreen
		if player.panim == PA_ABILITY
			fluffy.roll = $ + (roll - $)/4
		else
			fluffy.roll = 0
		end
		mo.rollangle = FixedMul(-fluffy.roll * ANG1/2, cos(R_PointToAngle(mo.x, mo.y) - (player.cmd.angleturn << 16)))
	end
	
	// super stuff
	
	if mo.state >= S_PLAY_SUPER_TRANS1
	and mo.state <= S_PLAY_SUPER_TRANS6
		// funny gauge go brrr
		if chrispy.supertrans < 30
		and player.charability == CA_FLY
		and player.cc_gaugeenabled ~= false
			local gauge = fluffy.supergauge
			if not valid(gauge)
				gauge = P_SpawnMobjFromMobj(mo, 0, 0, 8*FRACUNIT, MT_PARTICLE)
				gauge.state = S_FLUFFY_GAUGENEEDLE
				gauge.angle = mo.angle
				gauge.rollangle = ANGLE_180
				fluffy.supergauge = gauge
				
				local overlay = P_SpawnMobjFromMobj(gauge, 0, 0, 0, MT_OVERLAY)
				overlay.state = S_FLUFFY_GAUGE
				overlay.target = gauge
				
				S_StartSound(mo, sfx_appear)
			end
			
			gauge.tics = 2
			gauge.threshold = $ + 1
			gauge.rollangle = $ - gauge.threshold*ANG2
			
			P_TeleportMove(gauge,
				mo.x + P_RandomRange(-gauge.threshold, gauge.threshold) ^ 3,
				mo.y + P_RandomRange(-gauge.threshold, gauge.threshold) ^ 3,
				mo.z + P_RandomRange(-gauge.threshold, gauge.threshold) ^ 3 + flip*8*mo.scale)
			
			if mo.eflags & MFE_VERTICALFLIP
				gauge.z = $ + mo.height - gauge.height
			end
		end
	elseif valid(fluffy.supergauge)
		// Bust gauge
		local part = P_SpawnMobjFromMobj(mo, 0, 0, 8*FRACUNIT, MT_PARTICLE)
		part.momx = 3*part.scale
		part.momz = flip*4*part.scale
		part.flags = $ & ~MF_NOGRAVITY
		part.state = S_FLUFFY_GAUGENEEDLE
		part.fuse = 3*TICRATE
		part.rollangle = fluffy.supergauge.rollangle
		
		part = P_SpawnMobjFromMobj(mo, 0, 0, 8*FRACUNIT, MT_PARTICLE)
		part.momx = -3*part.scale
		part.momz = flip*4*part.scale
		part.flags = $ & ~MF_NOGRAVITY
		part.state = S_FLUFFY_GAUGE
		part.fuse = 3*TICRATE
		
		S_StartSound(mo, sfx_crumbl)
		P_RemoveMobj(fluffy.supergauge)
	elseif player.powers[pw_super]
	and gametyperules & GTR_FRIENDLY
		// spawn flickies
		for i = 1, NUMFLICKIES
			if not valid(fluffy.flickies[i])
				local bird = P_SpawnMobjFromMobj(mo, 0, 0, 256*FRACUNIT, MT_CHRISPY_FLICKY)
				bird.tracer = mo
				bird.target = mo
				bird.movedir = -i*ANGLE_90
				bird.color = FLICKY_COLORS[i]
				bird.sprite = FLICKY_SPRITES[i]
				bird.extravalue1 = bird.color
				bird.extravalue2 = bird.sprite
				bird.shadowscale = FRACUNIT
				fluffy.flickies[i] = bird
			end
		end
	end
	
	// Super flickies!
	
	if fluffy.flickydelay > 0 // bouncing
		fluffy.flickydelay = $ - 1
	elseif fluffy.flickies[1]
		local validflickies = {}
		
		for i = 1, NUMFLICKIES
			local bird = fluffy.flickies[i]
			if bird.valid
				table.insert(validflickies, bird)
			end
		end
		
		if not #validflickies
			fluffy.flickies = {}
		else
			// send flickies to attack!
			local radius = FixedMul(FLICKY_ATTACK_RADIUS, mo.scale)
			local bestitem
			local bestdist = INT32_MAX
			local necktarget = mo.neck and mo.neck.target
			
			flickytargets = {}
			searchBlockmap("objects", FlickySearch, mo,
				mo.x - radius, mo.x + radius,
				mo.y - radius, mo.y + radius)
			
			for i, item in ipairs(flickytargets)
				local gotonext
				
				for j, bird in ipairs(validflickies)
					if bird.target == item // we can assume validity here
						gotonext = true
						break
					end
				end
				
				if gotonext
					continue
				end
				
				if item == necktarget // prioritize anything Fluffy is looking at
					bestitem = necktarget
					break
				end
				
				local moz = mo.z + (mo.height >> 1)
				local itemz = item.z + (item.height >> 1)
				local dist = FixedHypot(FixedHypot(mo.x - item.x, mo.y - item.y), moz - itemz)
				
				if dist > radius
					continue
				end
				
				if dist < bestdist
					bestdist = dist
					bestitem = item
				end
			end
			
			if bestitem
				local bestbird // undoubtedly jart
				local bestdist = INT32_MAX
				local itemz = bestitem.z + (bestitem.height >> 1)
				
				for i, bird in ipairs(validflickies)
					if bird.fuse > 0 // bird is already attacking
					or bird.reactiontime > 0
						continue
					end
					
					local birdz = bird.z + (bird.height >> 1)
					local dist = FixedHypot(FixedHypot(bestitem.x - bird.x, bestitem.y - bird.y), itemz - birdz)
					
					if dist < bestdist
					and dist < radius
						bestbird = bird
						bestdist = dist
					end
				end
				
				if bestbird // this can be nil if all 4 birds are attacking or out of range
					local speed = FixedMul(bestbird.scale, bestbird.info.speed)
					bestbird.momx = FixedMul(FixedDiv(bestitem.x - bestbird.x, bestdist), speed)
					bestbird.momy = FixedMul(FixedDiv(bestitem.y - bestbird.y, bestdist), speed)
					bestbird.momz = FixedMul(FixedDiv(bestitem.z - bestbird.z, bestdist), speed)
					bestbird.fuse = FLICKY_ATTACK_FUSE
					bestbird.target = bestitem
					S_StartSound(bestbird, sfx_zoom)
					
					fluffy.flickydelay = FLICKY_ATTACK_INTERVAL
				end
			end
		end
	end
	
	// dab / hater
	
	if mo.state == S_CHRISPY_DAB
		player.drawangle = (player.cmd.angleturn << 16) - ANGLE_180 + ANG1
	elseif mo.state == S_CHRISPY_HATER
		player.drawangle = (player.cmd.angleturn << 16) - ANGLE_180
	end
	
	// Mace animation
	
	if player.powers[pw_carry] == CR_MACESPIN
		player.drawangle = R_PointToAngle2(0, 0, mo.momx, mo.momy)
	end
	
	// super visuals
	
	if player.powers[pw_super]
	and mo.state == S_PLAY_FLOAT
		mo.state = S_PLAY_WALK
	end
	
	// kill headtracking objects if no longer looking
	if looking and mo.state ~= S_PLAY_LOOK
		CC_KillNeck(mo.neck)
	end
	
	// resets
	fluffy.grounded = grounded
end

// thatstails Thinker
// I had to split this from the main thinker halfway through development for the thatstails feature so apologies if things get messy
// Some logic uses the camera angle for tail animation/placement variation and is thus NOT netgame synched, do NOT write gamelogic dependent on this unless you know what you're doing

local function TailsThinker(player, actor)
	local mo = player.mo
	actor = $ or mo
	
	if not mo
		return
	end
	
	// if the player is not meant to have tails, get rid of them!
	
	if not player.cc_thatstailsenabled
	and actor.skin ~= SKIN
	or player.powers[pw_carry] == CR_NIGHTSMODE
		if actor.fluffytails
			local tails = actor.fluffytails.tails
			for i = #tails, 1, -1
				local tail = tails[i]
				if valid(tail)
					P_RemoveMobj(tail)
				end
			end
			
			actor.fluffytails = nil
		end
		return
	end
	
	// if the player IS meant to have tails and they don't, set them up
	
	if not actor.fluffytails
		actor.fluffytails = {}
		local fluffy = actor.fluffytails
		
		fluffy.time = -1
		fluffy.momangle = R_PointToAngle2(0, 0, actor.momx - player.cmomx, actor.momy - player.cmomy)
		fluffy.state = mobjinfo[MT_FLUFFY_TAIL].spawnstate
		fluffy.tailoffset = 0
		fluffy.tailindex = 0
		fluffy.tails = {false, false} // table holding fluffy's tail objects (initialized to "false" so the script treats them as nonexistant)
		fluffy.frames = {}
		for i = 1, #fluffy.tails
			fluffy.frames[i] = 0
		end
	end
	
	if actor == mo
	and valid(player.followmobj) // fuck me, this is silly
	and player.followmobj.type <= MT_METALJETFUME
		P_RemoveMobj(player.followmobj)
	end
	
	// locals
	
	local chrispy = player.chrispy
	local fluffy = actor.fluffytails
	local flip = P_MobjFlip(actor)
	local grounded = P_IsObjectOnGround(actor)
	local slope = actor.standingslope
	
	fluffy.time = max(0, $ + 1)
	
	// flip direction on object bounce
	
	if actor == mo
		local momangle = R_PointToAngle2(0, 0, actor.momx - player.cmomx, actor.momy - player.cmomy) >> 16
		if abs(fluffy.momangle - momangle) > ANG1 >> 16
		and FixedHypot(actor.momx - player.cmomx, actor.momy - player.cmomy) > actor.scale
		and player.pflags & (PF_DIRECTIONCHAR|PF_JUMPED|PF_NOJUMPDAMAGE|PF_SPINNING) == (PF_DIRECTIONCHAR|PF_JUMPED)
		and not (player.cmd.forwardmove or player.cmd.sidemove)
			player.drawangle = momangle << 16
		end
		fluffy.momangle = momangle
	end
	
	// Tail animations!
	
	local camangle
	if splitscreen
		camangle = player.cmd.angleturn << 16
	elseif player.awayviewtics and valid(player.awayviewmobj)
		camangle = player.awayviewmobj.angle
	else
		camangle = R_PointToAngle(actor.x, actor.y)
	end
	local xcomponent = FixedMul(cos(camangle), actor.scale) // these offsets push the tails away from the camera just a tiny bit
	local ycomponent = FixedMul(sin(camangle), actor.scale) // this is so they are always placed behind his body if being viewed directly from the side
	
	local angle
	if mo == actor
		angle = player.drawangle
	else
		angle = actor.angle
	end
	
	local tics = 4 // duration that each tail is displayed for
	local offset = 2 // secondary tail animation is offset from primary
	local state // state to use for the tails
	local statecalls = {} // number of state calls the tails should perform
	local reset // if true, starts tails from the beginning of the animation
	local dist = 11 // number of FRACUNITS to move the tails behind the player
	local clientfactor = 1 // multiplier for how far back the tails should be pushed relative to the camera
	local zheight = 0 // number of FRACUNITS to move the tails up or down
	local noblur // set to true if ANIM_FLY or ANIM_RUN should not spawn the blur overlay
	local mirror = false // whether or not the tails should be mirrored
	local x, y, z
	
	local tracer = actor
	
	local sprite2 = actor.sprite2 & ~FF_SPR2SUPER
	local dead = sprite2 == SPR2_DEAD or spr2defaults[sprite2] == SPR2_DEAD
	
	/*if not (fluffy.linkmobj and fluffy.linkmobj.valid) // Fluffy's tails need to be render-linked to a separate object to use different colors from the player
		fluffy.linkmobj = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
		fluffy.linkmobj.tics = -1
	end
	P_TeleportMove(fluffy.linkmobj, // teleport the tails behind the player
				mo.x - dist*FixedMul(cos(angle), mo.scale) + clientfactor*xcomponent,
				mo.y - dist*FixedMul(sin(angle), mo.scale) + clientfactor*ycomponent,
				mo.z + zheight*flip*mo.scale)
	fluffy.linkmobj.scale = mo.scale
	fluffy.linkmobj.color = player.fluffy_primary or mo.color*/
	
	for i = 1, #fluffy.tails
		statecalls[i] = 0
	end
	
	// animation selection
	
	local anim = TAIL_ANIMS[actor.sprite] and (TAIL_ANIMS[actor.sprite][sprite2] or ANIM_TWIRL) or ANIM_NONE // pick an animation depending on the player's sprite and sprite2
	
	// switch animation under special cases
	
	if player.pflags & PF_STARTDASH
	and player.dashspeed >= player.maxdash
	or chrispy
	and (chrispy.flags & CF_DROPDASHCHARGE
	and chrispy.dropdash > 0
	or chrispy.fluffy and chrispy.fluffy.flightcharge > FLIGHTCHARGE_THRESHOLD)
		anim = ANIM_RUN
	end
	
	if sprite2 == SPR2_WALK
	and chrispy and chrispy.fluffy
	and (chrispy.fluffy.walktime > WALK_TIME/2 or chrispy.fluffy.mode == FM_DASH)
		anim = ANIM_RUN
		noblur = true
	end
	
	if sprite2 == SPR2_TIRE
	and actor.eflags & MFE_UNDERWATER
		anim = ANIM_TWIRL
	end
	
	if sprite2 == SPR2_CNT1
	and actor.skin == "fluffy"
		anim = ANIM_NONE
	end
	
	state = TAIL_STATES[anim]
	
	// more if-elses than yandere simulator
	if anim == ANIM_TWIRL // walk/roll
		if player.pflags & PF_STARTDASH
			local ratio = FixedDiv(player.dashspeed - player.mindash, (player.maxdash - player.mindash) or 1)
			tics = 3 - (3*ratio/FRACUNIT)
			offset = 2 + 2*ratio/FRACUNIT
		else
			local dohorizontal = true
			local docurrents = false
			local dovertical = false
			
			if sprite2 == SPR2_SPNG
			or sprite2 == SPR2_FALL
				dovertical = true
			elseif sprite2 == SPR2_FAN_
				dovertical = true
				docurrents = true
			elseif sprite2 == SPR2_SWIM
			or sprite2 == SPR2_TIRE
			or player.playerstate == PST_DEAD
				dohorizontal = false
			end
			
			if dohorizontal
				local speed
				if docurrents
					speed = FixedDiv(FixedDiv(FixedHypot(FixedHypot(actor.momx, actor.momy), actor.momz), mo.movefactor), actor.scale)
				else
					speed = FixedDiv(FixedDiv(FixedHypot(FixedHypot(actor.momx - player.cmomx, actor.momy - player.cmomy), actor.momz), mo.movefactor), actor.scale)
				end
				
				if sprite2 == SPR2_SPIN
				or speed > 52*FRACUNIT
					offset = 4
					tics = 1
				elseif speed > skins[actor.skin].runspeed
					offset = 4
					tics = 2
				elseif speed > 10*FRACUNIT
					offset = 3
					tics = 3
				else
					offset = 2
					tics = 4
				end
			end
			
			if dovertical
				local speed = FixedDiv(abs(actor.momz), actor.scale)
				if (speed < 10<<FRACBITS)
					tics = min($, 4)
				elseif (speed < 20<<FRACBITS)
					tics = min($, 3)
				elseif (speed < 30<<FRACBITS)
					tics = min($, 2)
				else
					tics = 1
				end
			end
				
		end
		
		if sprite2 == SPR2_ROLL
		or sprite2 == SPR2_TWIN
			dist = 24
			zheight = 4
		elseif sprite2 == SPR2_SWIM
			local d = P_GetPlayerControlDirection(player)
			
			if d
				offset = 4
				tics = 3 - d
			end
			dist = 14*cos(actor.rollangle)/FRACUNIT
			zheight = 10 + abs(14*sin(actor.rollangle)/FRACUNIT)
		elseif sprite2 == SPR2_TIRE
			state = S_FLUFFYTAIL_TWIRL_ABOVE
			dist = 8
			zheight = 12
		elseif sprite2 == SPR2_SPIN
			dist = 21
		elseif sprite2 == SPR2_FLIP
			dist = 18
			zheight = 10
		elseif sprite2 == SPR2_FAN_
			zheight = 12
			dist = 12
		elseif sprite2 == SPR2_HI5_
			dist = 9
			if camangle - player.drawangle < 0
				angle = camangle + ANGLE_112h
			else
				angle = camangle - ANGLE_112h
			end
		elseif sprite2 == SPR2_SPNG
		or sprite2 == SPR2_FALL
		or dead
			if dead
				dist = 0
				angle = camangle + ANGLE_180
			elseif sprite2 == SPR2_SPNG
				dist = 12
			else
				dist = 9
			end
			
			if actor.state ~= S_CHRISPY_TRICK
				local momz = flip*actor.momz/actor.scale
				zheight = min(max(-momz, -8), 8)
				if zheight > 2
					state = S_FLUFFYTAIL_TWIRL_ABOVE
				elseif zheight < -2
					state = S_FLUFFYTAIL_TWIRL_BELOW
				end
			end
		end
	elseif anim == ANIM_RUN // running
	or anim == ANIM_STOP
		offset = 4
		if sprite2 == SPR2_ROLL
		or sprite2 == SPR2_TWIN
			dist = 18
			zheight = 3
			tics = 2
		elseif sprite2 == SPR2_SPIN
		or sprite2 == SPR2_RUN
		or sprite2 == SPR2_SKID
			if sprite2 == SPR2_SPIN
				dist = 12
			else
				local viewdiff = abs(camangle - angle)
				if sprite2 == SPR2_SKID
					dist = 4
					if actor.state == S_FLUFFY_TOSS2
						state = S_FLUFFYTAIL_RUN
					end
				else
					dist = 7
				end
				if viewdiff < ANGLE_22h
					dist = $ + FixedMul(22, cos(viewdiff*4))
				end
			end
			tics = player.speed > 52*FixedMul(actor.scale, mo.movefactor) and 1 or 2
		else
			if player.panim == PA_EDGE
			or player.panim == PA_IDLE
				dist = 3
			else
				dist = 4
			end
			tics = 2
			zheight = -4
		end
	elseif anim == ANIM_FLY // flight
		dist = 1
		clientfactor = 0
		offset = 4
		if sprite2 == SPR2_TIRE
			tics = 3
			noblur = true
		else
			local rise = player.charflags & SF_MULTIABILITY and player.cmd.buttons & BT_JUMP
				or player.fly1
				or player.cmd.buttons & BT_USE
			tics = rise and 2 or 3
		end
	elseif anim == ANIM_HANG
		if sprite2 == SPR2_PAIN
			dist = 4
			zheight = -6
		else
			dist = 6
			zheight = -13
		end
	elseif anim == ANIM_DAB // dab
	or anim == ANIM_TPOSE
		offset = 0
		dist = 0
	elseif anim == ANIM_IDLE // idle
		if sprite2 == SPR2_LAND
			local charge = chrispy and chrispy.fluffy and chrispy.fluffy.flightcharge
			
			dist = 15
			
			if charge
				local thresh = 2*FLIGHTCHARGE_THRESHOLD/3
				if charge >= thresh
					state = S_FLUFFYTAIL_FLY
					dist = 8
					zheight = -5
					offset = 4
					tics = 2
					noblur = true
				else
					state = S_FLUFFYTAIL_TWIRL_ABOVE
					dist = 10
					zheight = 7
					tics = 3 - (3*charge / thresh)
					offset = 2 + (2*charge / thresh)
				end
			end
		elseif sprite2 == SPR2_EDGE
			dist = 10
			offset = 4
			if actor.state == S_CHRISPY_ACT
			and chrispy
				tics = 3 + (chrispy.emotewheel.tics % 3)
			else
				tics = 3 + (actor.tics % 3)
			end
			mirror = ((actor.frame & FF_FRAMEMASK) & 1 and true or false)
			if fluffy.prevsprite2 ~= SPR2_EDGE
			or fluffy.prevmirrored ~= actor.mirrored
				reset = true
			end
		elseif sprite2 == SPR2_DNCE
			if player.cc_dance
				tics = (player.cc_dance.tics & 1) + 2
			end
			state = S_FLUFFYTAIL_TWIRL_MIDDLE
			angle = camangle + ANGLE_180
		elseif sprite2 == SPR2_TRNS|FF_SPR2SUPER
			angle = camangle - ANGLE_135
		elseif sprite2 == SPR2_CNT1
			angle = camangle
		end
	end
	
	if state == S_FLUFFYTAIL_TWIRL
		if (fluffy.state == S_FLUFFYTAIL_TWIRL_ABOVE or fluffy.state == S_FLUFFYTAIL_TWIRL_BELOW)
			fluffy.middle = true
		end
		
		if fluffy.middle
			if player.panim == PA_WALK
			or abs(camangle - angle) > ANGLE_45
				fluffy.middle = false
			else
				state = S_FLUFFYTAIL_TWIRL_MIDDLE
			end
		end
	else
		fluffy.middle = false
	end
		
	// choose position
	
	if anim == ANIM_TWIRL
	and (sprite2 == SPR2_ROLL or sprite2 == SPR2_TWIN or sprite2 == SPR2_FLIP)
		local factor = player.powers[pw_carry] == CR_MACESPIN and 1 or 2
		local hspeed = max(FixedHypot(actor.momx, actor.momy), factor*actor.scale)
		local vangle = sprite2 == SPR2_FLIP and (actor.frame & FF_FRAMEMASK + 1) * FixedAngle(360*FRACUNIT/7)
			or slope and FixedMul(cos(angle - slope.xydirection), slope.zangle)
			or R_PointToAngle2(0, 0, hspeed, factor*actor.momz)
			
		local trans = chrispy and chrispy.ddtrans
		if trans
			if trans > 0 // transition from dropdash to normal pos
				vangle = $ + trans * (-$ / DROPDASH_DELAY)
			elseif trans < 0 // transition from position to dropdash
				trans = abs($)
				vangle = trans * ($ / DROPDASH_DELAY)
			end
		end
		
		local hdist = FixedMul(cos(vangle), actor.scale)
		
		local thresholdangle = ANGLE_45
		
		if flip*vangle < -thresholdangle
			state = S_FLUFFYTAIL_TWIRL_ABOVE
		elseif flip*vangle > thresholdangle
			state = S_FLUFFYTAIL_TWIRL_BELOW
		end
		
		x = actor.x - dist*FixedMul(hdist, cos(angle)) + clientfactor*xcomponent
		y = actor.y - dist*FixedMul(hdist, sin(angle)) + clientfactor*ycomponent
		z = actor.z - dist*FixedMul(sin(vangle), actor.scale) + zheight*flip*actor.scale
	else
		x = actor.x - dist*FixedMul(cos(angle), actor.scale) + clientfactor*xcomponent
		y = actor.y - dist*FixedMul(sin(angle), actor.scale) + clientfactor*ycomponent
		z = actor.z + zheight*flip*actor.scale
	end
	
	// animation handling
	
	if states[state].frame ~= fluffy.stateframe
	or reset
		fluffy.stateframe = states[state].frame
		fluffy.tailtics = 0 // these values control which tail is displayed and for how long
		if states[state].var2 and states[fluffy.state].var2
			for i = 1, #fluffy.frames
				statecalls[i] = $ + fluffy.frames[i]
			end
		else
			fluffy.tailoffset = offset
			statecalls[1] = $ + offset
		end
	end
	
	if fluffy.tailtics
		fluffy.tailtics = $ - 1
	end
	
	if fluffy.tailtics <= 0
		if fluffy.tailoffset > offset
			fluffy.tailoffset = $ - 1
			statecalls[2] = $ + 1
		elseif fluffy.tailoffset < offset
			fluffy.tailoffset = $ + 1
			statecalls[1] = $ + 1
		end
		
		fluffy.tailtics = tics
		for i = 1, #statecalls
			statecalls[i] = $ + 1
		end
	end
	
	fluffy.prevsprite2 = actor.sprite2
	fluffy.prevmirrored = actor.mirrored
	fluffy.state = state
	
	// set tracer to head if neck is active
	if actor.neck
		local dir = R_PointToAngle2(x, y, actor.x, actor.y)
		local linkhead = actor.neck.heads[2]
		if valid(linkhead)
		and abs(camangle - dir) <= ANGLE_90
			tracer = linkhead
		end
	end
	
	// handle the tails and apply animation
	
	for i = #fluffy.tails, 1, -1 // iterate through all the tails (should be just 2)
		if not (fluffy.tails[i] and fluffy.tails[i].valid) // spawn in a tail if it does not exist!
			local tail = P_SpawnMobj(actor.x, actor.y, actor.z, MT_FLUFFY_TAIL)
			tail.target = actor
			tail.skin = SKIN
			if actor.info.dispoffset >= 0
				tail.flags2 = $ | MF2_LINKDRAW
			end
			fluffy.tails[i] = tail
		end
		
		local tail = fluffy.tails[i] // shortcut to access the tail
		local x = x
		local y = y
		
		if tail.tracer ~= tracer // tracer can be either the player or the player's head, if they're a neck user
			tail.tracer = tracer
		end
		
		// sync color, scale, flags with player
		tail.fuse = 2 // ez removal
		tail.color = actor.color//player.fluffy_primary or mo.color
		tail.scale = actor.scale
		tail.height = actor.height
		tail.eflags = actor.eflags
		tail.angle = angle
		tail.mirrored = mirror ~= (states[state].var1 and true or false)
		
		// special case bruh moments
		
		if i == 1
			if anim == ANIM_DAB
				x = $ - 2*clientfactor*xcomponent
				y = $ - 2*clientfactor*ycomponent
			end
			
			if CC_Renderer.value == 2
				x = $ - xcomponent/4
				y = $ - ycomponent/4
			end
		end
		
		P_TeleportMove(tail, x, y, z)
		
		// pick a tail frame to be displayed!
		if i ~= 1
			tail.sprite2 = $ & ~FF_SPR2SUPER
		end
		if reset
			tail.state = S_INVISIBLE
		end
		while statecalls[i]
			statecalls[i] = $ - 1
			tail.state = state
		end
		if i ~= 1
			tail.sprite2 = $ | FF_SPR2SUPER
		end
		fluffy.frames[i] = tail.frame & FF_FRAMEMASK
		tail.frame = fluffy.frames[i] | (actor.frame & ~FF_FRAMEMASK)
		
		// flicker overlay
		if i == 1
		and (state == TAIL_STATES[ANIM_RUN] or state == TAIL_STATES[ANIM_FLY])
		and (tail.frame & FF_FRAMEMASK) & 1 == 0
		and not (actor.flags2 & MF2_DONTDRAW)
		and not noblur
		and player.cc_flickerenabled ~= false
			local ghost = P_SpawnMobjFromMobj(tail, 0, 0, 0, MT_FLUFFY_TAIL)
			ghost.tracer = tracer
			ghost.flags2 = $ | (tail.flags2 & MF2_LINKDRAW)
			ghost.fuse = 1
			ghost.angle = tail.angle
			ghost.color = tail.color
			ghost.colorized = tail.colorized
			ghost.skin = SKIN
			ghost.state = S_FLUFFYTAIL_BLUR
			ghost.frame = (state == TAIL_STATES[ANIM_RUN] and B or A) | (tail.frame & ~(FF_FRAMEMASK|FF_TRANSMASK)) | FF_TRANS50
		end
	end
end

rawset(_G, "CC_TailsThinker", TailsThinker) // for external use, such as in SRB2P

local function CarryThinker(player1)
	local mo1 = player1.mo
	if not mo1 return end
	
	local mo2 = mo1.tracer
	local player2 = valid(mo2) and mo2.player
	local chrispy2 = player2 and player2.chrispy
	local fluffy2 = chrispy2 and chrispy2.fluffy
	
	if not fluffy2 return end // don't run for non-Fluffy carriers
	
	if player1.powers[pw_carry] == CR_PLAYER
	and player2.powers[pw_tailsfly]
	and player2.panim == PA_ABILITY
	and not (chrispy2.flags & CF_TOSSING)
		local target = P_LookForEnemies(player2, false)
		local crouchbutton = player2.cc_buttonmap and player2.cc_buttonmap.crouch or CC_DEFAULTBUTTONS.crouch
		if chrispy2.buttons[crouchbutton]
		or (player1.chrispy and player1.chrispy.buttons[crouchbutton] == 1)
			fluffy2.target = target or mo1
			fluffy2.cargo = mo1
			fluffy2.tossangle = player2.drawangle
			chrispy2.flags = $ | CF_TOSSING
			mo2.state = S_FLUFFY_TOSS1
			S_StartSound(mo2, sfx_s236)
			return
		end
		if target
			P_SpawnLockOn(player2.bot and player1 or player2, target, S_LOCKON1)
		end
		mo2.state = S_FLUFFY_CARRY
	end
end

addHook("ThinkFrame", do
	for player in players.iterate
		FluffyThinker(player)
		CarryThinker(player)
	end
end)

addHook("PostThinkFrame", do
	for player in players.iterate
		TailsThinker(player)
	end
end)

// Super flickies!

local function FlickyFuse(bird)
	bird.target = nil
	bird.fuse = 0
	return true
end

addHook("MobjFuse", FlickyFuse, MT_CHRISPY_FLICKY)

addHook("MobjMoveCollide", function(bird, item)
	if not (bird.valid and item.valid)
		return
	end
	
	local player = bird.tracer.valid and bird.tracer.player
	
	if item == bird.target
	and bird.z <= item.z + item.height
	and item.z <= bird.z + bird.height
		P_DamageMobj(item, bird, bird.tracer)
		FlickyFuse(bird)
		bird.reactiontime = 16
		P_SetObjectMomZ(bird, 10*FRACUNIT)
		P_InstaThrust(bird, bird.angle, -10*FRACUNIT)
		if not (player and player.powers[pw_super])
			bird.lastlook = 1
		end
	end
end, MT_CHRISPY_FLICKY)

addHook("MobjThinker", function(bird)
	local mo = bird.tracer
	local player = valid(mo) and mo.player
	
	// no player? fuck that
	if not player
		P_RemoveMobj(bird)
		return
	end
	
	// if the player changes skin, fly away
	if mo.skin ~= SKIN
		bird.lastlook = 2
	elseif bird.lastlook == 2 // if player switch back to Fluffy, despawn - Fluffy will spawn new ones in, so let's avoid bird spam
		P_RemoveMobj(bird)
		return
	end
	
	local super = player.powers[pw_super]
	local exit = bird.lastlook > 0
	
	// if flying away, despawn once you're in the floor/ceiling
	if exit
	and (bird.z > bird.ceilingz + 256*bird.scale
		or bird.z < bird.floorz - 256*bird.scale)
			P_RemoveMobj(bird)
			return
	end
	
	// match player
	bird.flags2 = $ & ~MF2_OBJECTFLIP | (mo.flags2 & MF2_OBJECTFLIP)
	bird.eflags = $ & ~MFE_VERTICALFLIP | (mo.eflags & MFE_VERTICALFLIP)
	bird.scale = mo.scale
	
	// color
	if player.pflags & PF_GODMODE
	and super
		bird.color = SKINCOLOR_SUPERGOLD1 + abs(((super >> 1) % 9) - 4)
	else
		bird.color = bird.extravalue1
	end
	
	// solidity
	local target = bird.target
	
	if target
		if not target.valid
		or not target.health
		or target.flags2 & MF2_FRET
			FlickyFuse(bird)
			target = mo
		end
	else
		target = mo
	end
	
	if exit
		bird.flags = $ | (MF_NOCLIP|MF_NOCLIPHEIGHT) // clip through everything while flying away
	elseif bird.flags & MF_NOCLIP // currently in a wall, once the flicky is out they should be made tangible again
		if bird.z > bird.floorz
		and bird.z + bird.height < bird.ceilingz
		and P_CheckSight(bird, target)
			bird.flags = $ & ~(MF_NOCLIP|MF_NOCLIPHEIGHT)
		end
	elseif not P_CheckSight(bird, target) // can't see their target? clip through walls until they do
		bird.flags = $ | (MF_NOCLIP|MF_NOCLIPHEIGHT)
	end
	
	// trail
	if target ~= mo
		local ghost = P_SpawnGhostMobj(bird)
		ghost.colorized = true
		if super
			ghost.frame = $ | FF_FULLBRIGHT
		end
	elseif super and leveltime & 1
		local star = P_SpawnMobjFromMobj(bird, 0, 0, 6*FRACUNIT, MT_CHRISPY_SPARKLE)
		star.color = bird.color
		star.colorized = true
	end
	
	// movement
	if bird.reactiontime > 0
		bird.reactiontime = $ - 1
		if bird.state ~= bird.info.seestate
			bird.state = bird.info.seestate
		end
	else
		local accel_h = FixedMul(FLICKY_ACCEL_H, mo.scale)
		local accel_z = FixedMul(FLICKY_ACCEL_Z, mo.scale)
		
		local goal = {}
		if exit
			goal.x = bird.x
			goal.y = bird.y
			if bird.eflags & MFE_VERTICALFLIP
				goal.z = bird.floorz - 256*FRACUNIT
			else
				goal.z = bird.ceilingz + 256*FRACUNIT
			end
		elseif target == mo
			local n = mo.chrispy and mo.chrispy.time or 0
			local ang_h = 2*n*ANG2 + bird.movedir
			local ang_z = n*ANG10 + 2*bird.movedir
			local radius = 3*mo.radius
			goal.x = target.x + P_ReturnThrustX(bird, ang_h, radius)
			goal.y = target.y + P_ReturnThrustY(bird, ang_h, radius)
			goal.z = target.z
				+ (mo.eflags & MFE_VERTICALFLIP and -bird.height - 9*bird.scale or target.height + 9*bird.scale)
				+ FixedMul(sin(ang_z), 8*bird.scale)
		else
			goal.x = target.x
			goal.y = target.y
			goal.z = target.z + (target.height >> 1) - (bird.height >> 1)
		end
		
		bird.angle = R_PointToAngle2(0, 0, bird.momx, bird.momy)
		
		for i, var in ipairs({"z", "y", "x"})
			local accel = i == 1 and target == mo and accel_z or accel_h
			local mom = "mom"..var
			local displacement = bird[var] - goal[var]
			if abs(displacement) < 6*accel
			and abs(bird[mom]) < 6*accel
				bird[mom] = -displacement/2
			elseif displacement > 0
				if bird[mom] > 0
					bird[mom] = $ - 3*accel
				else
					bird[mom] = $ - accel
				end
			else
				if bird[mom] < 0
					bird[mom] = $ + 3*accel
				else
					bird[mom] = $ + accel
				end
			end
		end
		
		// state selection
		if target == mo
		and P_MobjFlip(bird)*bird.momz < -10*bird.scale
			if bird.state ~= bird.info.seestate
				bird.state = bird.info.seestate
			end
		else
			if bird.state ~= bird.info.spawnstate
				bird.state = bird.info.spawnstate
			end
		end
	end
end, MT_CHRISPY_FLICKY)

// Fluffy shouldn't damage things while super
// These conditions are NOT exhaustive, they really only account for generic SP stuff - if you encounter a situation I need to address, post in the MB thread!

local function FluffyIgnoreDamage(player)
	local mo = player.mo
	
	return not player.powers[pw_super]
	or not (gametyperules & GTR_FRIENDLY)
	or mo.skin ~= SKIN
	or mo.state >= S_PLAY_SUPER_TRANS1 and mo.state <= S_PLAY_SUPER_TRANS6
	or player.powers[pw_invulnerability]
	or player.pflags & PF_GODMODE
	or player.charflags ~= skins[mo.skin].flags
	or player.charability ~= skins[mo.skin].ability
	or player.charability2 and player.charability2 ~= skins[mo.skin].ability2
end

addHook("PlayerCanDamage", function(player, item)
	local mo = player.mo
	
	if player.pflags & (PF_JUMPED|PF_NOJUMPDAMAGE) == PF_JUMPED
	or player.pflags & PF_SPINNING
	or FluffyIgnoreDamage(player)
	or not item.valid
		return
	end
	
	local flip = P_MobjFlip(mo)
	local bottomheight = mo.z
	local topheight = mo.z + mo.height
	
	if mo.eflags & MFE_VERTICALFLIP
		local swap = bottomheight
		bottomheight = topheight
		topheight = swap
	end
	
	if flip*(topheight - (item.z + (item.height >> 1))) < 0
	and player.panim == PA_ABILITY and not (mo.eflags & MFE_UNDERWATER)
	and flip*(mo.momz - item.momz) > 0
		return
	end
	
	return false
end)

addHook("ShouldDamage", function(mo, item, source, damage, damagetype)
	local player = mo.player
	if not player
	or FluffyIgnoreDamage(player)
	//or player.exiting // active super should make this redundant
	or damagetype & DMG_DEATHMASK
	or G_PlatformGametype() and item and source and source.player
	or player.powers[pw_flashing] or player.powers[pw_invulnerability]
		return
	end
	
	P_DoPlayerPain(player, item, source)
	if item and item.info.mass == DMG_SPIKE
		S_StartSound(mo, sfx_spkdth)
	else
		S_StartSound(mo, sfx_shldls)
	end
	return
	
end, MT_PLAYER)

// cache fly and swim frames to reduce lag during gameplay
// not necessary in 2.2.9+

if not (MODID == 18 and SUBVERSION < 9)
	return
end

local cachepatch

local function CacheFluffy(v)
	if not valid(cachepatch)
		local skin = SKIN
		local sprites = {SPR2_FLY, SPR2_SWIM}
		local spr2names = {"FLY_", "SWIM"}
		
		for i, spr2 in ipairs(sprites)
			for ang = 1, 8
				local frame = 0
				while v.getSprite2Patch(skin, spr2, false, frame, ang, ANGLE_45)
					print("Rotated patches for "..spr2names[i]..R_Frame2Char(frame)..ang.." cached for skin "..skin..".")
					frame = $ + 1
				end
			end
		end
		cachepatch = v.cachePatch("MISSING")
	end
end

hud.add(CacheFluffy, "title")
hud.add(CacheFluffy, "titlecard")