/*
Pointy Sonic
	by Lach

beta
*/

local SKIN = "pointy" // skin name to run this script for

local TEETER_MIN_TICS = 3 // minimum and maximum duration for randomized teeter frames
local TEETER_MAX_TICS = 6

local PEELOUT_SPEED = 50*FRACUNIT // the speed of a peelout launch
local SPEED_INCREASE_LEEWAY = 3*FRACUNIT // the amount of speed above normalspeed needed to update normalspeed
local SPEED_DECREASE_LEEWAY = 8*FRACUNIT // the amount of speed below normalspeed needed to update normalspeed
local PEELOUT_GRAB_TIME = 8 // amount of time after releasing a peelout during which Pointy can grab another player

local TPOSEDEATH_TIME = 8*60*TICRATE // amount of time player must stay in WAIT before t-posing
local TPOSEDEATH_HOLD = TICRATE // amount of time to hold tpose before dying

rawset(_G, "CC_SPEED_INCREASE_LEEWAY", SPEED_INCREASE_LEEWAY) // needed in CHRISPYCHARS.lua

local function valid(mo)
	return mo and mo.valid
end

local function Squiggle(emote)
	P_TeleportMove(emote,
		emote.x + P_RandomRange(-1, 1)*emote.scale,
		emote.y + P_RandomRange(-1, 1)*emote.scale,
		emote.z + P_RandomRange(-1, 1)*emote.scale)
end

local function Pulsate(emote, item, mo, rate)
	rate = $ or 12
	emote.z = $ + 4*P_MobjFlip(emote)*mo.scale
	emote.threshold = ($ + 1) % rate
	if emote.threshold < 6
		emote.scale = 3*$/2
		emote.z = $ - P_MobjFlip(emote)*4*emote.scale
	end
end

local function Bob(emote)
	emote.threshold = ($ + 1) % 72
	emote.z = $ + 3*emote.scale + 3*FixedMul(sin(emote.threshold*(ANG10/2)), emote.scale)
end

local function Exclamation(emote, item, mo)
	Pulsate(emote, item, mo)
	if emote.threshold < 6
		if item.flags & MF_BOSS
			emote.color = SKINCOLOR_CRIMSON
		elseif item.player
			emote.color = item.player.skincolor
		else
			emote.color = SKINCOLOR_RED
		end
	end
end

local function Question(emote)
	emote.threshold = ($ + 1) % 36
	emote.z = $ + 6*P_MobjFlip(emote)*FixedMul(abs(sin(emote.threshold*ANG10)), emote.scale)
	if emote.cusval
		emote.rollangle = FixedMul(ANG30, cos(emote.threshold*ANG10))
	elseif emote.threshold < 18
		emote.rollangle = FixedMul((36 - emote.threshold)*FRACUNIT >> 2, (2*(emote.threshold/2 & 1) - 1)*ANG1)
	end
end

local function Sneeze(emote, item, mo)
	local EMOTEDELAY = 5*TICRATE/3 // countdown until emote appears
	local SNEEZETHRESHOLD = 2*TICRATE/5 // countdown until sneeze
	
	if emote.movedir ~= emote.lastlook
		emote.threshold = -EMOTEDELAY
	end
	
	if emote.threshold >= 0
		Question(emote)
		return
	end
	
	emote.state = S_INVISIBLE
	
	if mo.tics < CC_NECK_TRANSITIONTIME
		return
	end
	
	emote.threshold = $ + 1
	if emote.threshold == -(EMOTEDELAY - SNEEZETHRESHOLD)
		local head = mo.neck.heads[2]
		if valid(head)
			head.state = S_POINTY_SNEEZE1
			S_StartSound(mo, sfx_sneeze)
		end
	elseif emote.threshold == 0
		S_StartSound(mo, sfx_emot2)
	end
end

local function Icon(emote, item)
	local z = emote.z
	local flip = P_MobjFlip(emote)
	emote.threshold = ($ + 1) % 72
	emote.z = $ + 6*flip*FixedMul(abs(sin(emote.threshold*ANG10/2)), emote.scale)
	emote.scale = $ + flip*(emote.z - z)/64
end

local function PlayerIcon(emote, item)
	if item.player
		if emote.skin ~= item.skin
			emote.skin = item.skin
		end
		emote.color = item.player.skincolor
	else
		emote.color = skins[emote.skin].prefcolor
	end
	emote.state = $
	emote.z = $ + P_MobjFlip(emote)*4*emote.scale
	Icon(emote, item)
end

local function Giggle(emote, item)
	emote.threshold = ($ + 1) % 6
	if emote.threshold < 3
		emote.z = $ + P_MobjFlip(emote)*2*emote.scale
	end
end

local function Hater(emote, item, mo)
	emote.threshold = ($ + 1) % TICRATE
	if emote.threshold < TICRATE/3
		Squiggle(emote)
	end
	if mo.player
		emote.color = mo.player.skincolor
	end
end

local function OkHand(emote, item, mo)
	local ang
	local flip = P_MobjFlip(emote)
	emote.z = $ + flip*3*mo.scale
	emote.threshold = ($ + 1) % (2*TICRATE)
	ang = FixedAngle(360*(emote.threshold % TICRATE)*FRACUNIT/TICRATE)
	
	if emote.threshold < TICRATE
		emote.rollangle = ang
	else
		local ogscale = emote.scale
		emote.scale = $ + abs(sin(ang))/2
		emote.z = $ - P_MobjFlip(emote)*10*(emote.scale - ogscale)
	end
end

local function DabParticles(emote, item, mo)
	emote.threshold = ($ + 1) % 40
	if emote.threshold % 10 == 0
		local n = emote.threshold / 10 % 4
		local colors = {SKINCOLOR_SUPERGOLD5, SKINCOLOR_ROSY, SKINCOLOR_NONE, SKINCOLOR_SKY}
		local part = P_SpawnMobjFromMobj(emote, 0, 0, 0, MT_PARTICLE)
		part.momz = P_MobjFlip(part)*mo.scale
		part.momx = P_RandomRange(-1, 1)*mo.scale
		part.momy = P_RandomRange(-1, 1)*mo.scale
		part.scale = 0
		part.destscale = mo.scale
		part.scalespeed = part.destscale/6
		part.fuse = 16
		part.sprite = SPR_EMOT
		part.frame = n + 2
		part.color = colors[n + 1]
	end
end

local function Surprised(emote, item, mo)
	Pulsate(emote, item, mo, 10)
	if mo.skin == SKIN
		emote.color = mo.player.skincolor
	else
		emote.color = skins[SKIN].prefcolor
	end
end

local CLAPTIME = 6
local function Clap(emote, item, mo)
	local dance = item.cc_dance
	
	if not dance return end
	
	if emote.movedir ~= emote.lastlook
	or emote.danceskin ~= dance.skin // yeah, I have to cheat a bit...
		local info = CC_DanceSkins[dance.skin]
		local time = 0
		local keys = info[dance.num].keys
		for i = 1, #keys
			time = $ + keys[i].tics
		end
		time = $ / (2*CLAPTIME) + 1
		emote.threshold = (#keys / time) << 16
		emote.danceskin = dance.skin
	end
	
	if (dance.key - 1) % (emote.threshold >> 16) == 0//((emote.threshold >> 8) & 7)
	and emote.threshold & 7 == 0
		S_StartSound(mo, sfx_emot5)
		emote.threshold = $ + CLAPTIME
		//emote.threshold = $ & ~(7 << 8) + (P_RandomKey(2) << 8)
	end
	
	if emote.threshold & 7
		emote.threshold = $ - 1
		emote.frame = $ + 1
	end
end	

local function Cash(emote)
	emote.threshold = ($ + 1) % 40
	if emote.threshold < 30
		emote.scale = $ + (sin(emote.threshold*ANG1*6) >> 1)
		emote.rollangle = FixedMul(sin(emote.threshold*ANG30), ANG15)
	end
end

if not CC_NeckSkins
	rawset(_G, "CC_NeckSkins", {})
end
CC_NeckSkins[SKIN] = { // Neck support
	skin = SKIN,
	speed = FRACUNIT/5,
	pivot = true,
	pitches = 5,
	idleangle = -ANGLE_45,
	superidleangle = 0,
	superidlepitch = -ANGLE_45,
	superpivot = false,
	superpitches = 3,
	superpitchinterval = ANG20*2,
	superstand = true,
	superfloatthreshold = 20*FRACUNIT,
	eastereggname = "\x70\048\x31\110\x74\032\x65\045\x73\111\x6e\105\x63",
	trackables = {
		{ // cash confetti when spawned via emote wheel
			condition = function(item, player)
				return item.state == S_CHRISPY_CASH
				and item.flags2 & MF2_SKULLFLY
				and item.target == player.mo
			end,
			emotesprite = SPR_EMOT,
			emoteframe = M, // $ Fang
			emotecolor = SKINCOLOR_DREAMY,
			emotesound = sfx_mario4,
			emotethink = Cash,
			fangmoney = true // identifier for emote wheel
		},
		{ // rings/coins when none are possessed
			condition = function(item, player)
				return player.rings < 1 and not player.bot
				and (item.type == MT_RING or item.type == MT_FLINGRING
					or item.type == MT_COIN or item.type == MT_FLINGCOIN
					or item.type == MT_REDTEAMRING and player.ctfteam == 1
					or item.type == MT_BLUETEAMRING and player.ctfteam == 2)
			end,
			emotesprite = SPR_EMOT,
			emoteframe = A, // squiggle
			emotecolor = SKINCOLOR_MAGENTA,
			emotesound = sfx_emot1,
			emotethink = Squiggle,
			squiggle = true // identifier for emote wheel
		},
		{ // air bubbles when underwater
			condition = function(item, player)
				return item.type == MT_EXTRALARGEBUBBLE
				and player.powers[pw_underwater] > 0 and player.powers[pw_underwater] < 20*TICRATE end,
			emotesprite = SPR_EMOT,
			emoteframe = A, // squiggle
			emotecolor = SKINCOLOR_MAGENTA,
			emotesound = sfx_emot1,
			emotethink = Squiggle
		},
		{ // ACZ3 Fang KO
			condition = function(item) return item.state == S_FANG_KO and item.type == MT_FANG end,
			emotesprite = SPR_EMOT,
			emoteframe = F, // chaojoy
			emotecolor = SKINCOLOR_SKY,
			emotesound = sfx_ghosty,
			emotethink = Giggle,
			ignorehealthcheck = true,
			chaojoy = true // identifier for emote wheel
		},
		{ // ACZ3 Fang
			condition = function(item) return item.type == MT_FANG and item.state ~= S_FANG_INTRO12 end,
			emoteskin = "fang",
			emotestate = S_PLAY_ICON1,
			emotesound = sfx_emot3,
			emotethink = PlayerIcon,
			ignorehealthcheck = true
		},
		{ // BCZ Metal Sonic
			condition = function(item) return item.type == MT_METALSONIC_RACE or item.type == MT_METALSONIC_BATTLE end,
			emoteskin = "metalsonic",
			emotestate = S_PLAY_ICON1,
			emotesound = sfx_emot3,
			emotethink = PlayerIcon
		},
		{ // FHZ Amy
			condition = function(item) return item.type == MT_ROSY end,
			emoteskin = "amy",
			emotestate = S_PLAY_ICON1,
			emotesound = sfx_emot4,
			emotethink = PlayerIcon
		},
		{ // Eggman bosses
			condition = function(item)
				return item.type == MT_EGGMOBILE
				or item.type == MT_EGGMOBILE2
				or item.type == MT_EGGMOBILE3
				or item.type == MT_EGGMOBILE4
				or item.type == MT_BLACKEGGMAN
				or item.type == MT_CYBRAKDEMON
			end,
			emotesprite = SPR_EMOT,
			emoteframe = J, // eggman face
			emotesound = sfx_emot3,
			emotethink = Icon,
			ignorehealthcheck = true,
			eggman = true // identifier for emote wheel
		},
		{ // bosses & enemy players
			condition = function(item, player)
				return item.flags & MF_BOSS
				or item.type == MT_FAKEMOBILE
				or item.player and not item.player.spectator
				and (item.player.ctfteam ~= player.ctfteam
					or G_RingSlingerGametype()
					and (not (gametyperules & (GTR_TAG|GTR_TEAMS)) or player.pflags & PF_TAGIT ~= item.player.pflags & PF_TAGIT))
			end,
			emotesprite = SPR_EMOT,
			emoteframe = C, // !
			emotecolor = SKINCOLOR_SUPERGOLD5,
			emotesound = sfx_emot3,
			emotethink = Exclamation,
			exclamation = true // identifier for emote wheel
		},
		{ // enemies/hazards
			condition = function(item, player)
				if item.flags & (MF_ENEMY|MF_PAIN|MF_FIRE|MF_MISSILE) or item.type == MT_EGGROBO1
					local dmgtoshield = {
						[DMG_WATER] = SH_PROTECTWATER,
						[DMG_FIRE] = SH_PROTECTFIRE,
						[DMG_ELECTRIC] = SH_PROTECTELECTRIC,
						[DMG_SPIKE] = SH_PROTECTSPIKE
					}
					if item.flags & MF_SCENERY
					or item.flags & MF_MISSILE and item.target == player.mo
					or item.flags & MF_FIRE and player.powers[pw_shield] & SH_PROTECTFIRE
					or player.powers[pw_shield] & (dmgtoshield[item.info.mass & ~(8|16|32|64)] or 0)
						return false
					end
					return true
				end
			end,
			emotesprite = SPR_EMOT,
			emoteframe = C, // !
			emotecolor = SKINCOLOR_SUPERGOLD5,
			emotesound = sfx_emot3,
			emotethink = Exclamation
		},
		{ // high-priority sneezables
			condition = function(item, player) return item.type == MT_SUSPICIOUSFACESTABBERSTATUE and not player.powers[pw_super] end,
			emotesprite = SPR_EMOT,
			emoteframe = B,
			emotecolor = SKINCOLOR_SUPERGOLD5,
			emotethink = Sneeze,
			cusval = 1,
			sneeze = true // identifier for Fluffy Tails
		},
		{ // highscore money
			condition = function(item) return item.state == S_CHRISPY_CASH or item.state == S_CHRISPY_COIN end,
			emotesprite = SPR_EMOT,
			emoteframe = M, // $ Fang
			emotecolor = SKINCOLOR_DREAMY,
			emotesound = sfx_mario4,
			emotethink = Cash
		},
		{ // dancing players
			condition = function(item) return item.state == S_CHRISPY_DANCE and item.player end,
			emotesprite = SPR_EMOT,
			emoteframe = G,
			emotethink = Clap
		},
		{ // dabbing players, if you're a dabber
			condition = function(item, player)
				return item.state == S_CHRISPY_DAB
				and not player.cc_haterenabled
			end,
			emotesprite = SPR_NULL,
			emotesound = sfx_emot6,
			emotethink = DabParticles
		},
		{ // hating players, if you're a hater
			condition = function(item, player)
				return item.state == S_CHRISPY_HATER
				and player.cc_haterenabled
			end,
			emotesprite = SPR_EMOT,
			emoteframe = E, // ok hand
			emotesound = sfx_emot0,
			emotethink = OkHand,
			okhand = true // identifier for emote wheel
		},
		{ // dabbing players, if you're a hater, or hating players, if you're a dabber
			condition = function(item, player)
				return item.state == S_CHRISPY_DAB
				and player.cc_haterenabled
				or item.state == S_CHRISPY_HATER
				and not player.cc_haterenabled
			end,
			emotesprite = SPR_EMOT,
			emoteframe = I, // shadow
			emotecolor = SKINCOLOR_RED,
			emotesound = sfx_emot1,
			emotethink = Hater,
			hater = true // identifier for hater feature/emote wheel
		},
		{ // speedy players
			condition = function(item, player)
				return item.player and item.player.speed > 48*player.mo.scale
			end,
			emotesprite = SPR_EMOT,
			emoteframe = K, // surprised pikachu
			//emotecolor = skins[SKIN].prefcolor, // GRRRR fuck lump load order
			emotecolor = SKINCOLOR_SAPPHIRE,
			emotesound = sfx_emot3,
			emotethink = Surprised,
			surprised = true // identifier for emote wheel
		},
		{ // t-posing players
			condition = function(item, player)
				return item.player and item.sprite2 == SPR2_TPOS and item.sprite == SPR_PLAY
			end,
			emotesprite = SPR_EMOT,
			emoteframe = O,
			emotecolor = SKINCOLOR_SAPPHIRE,
			emotesound = sfx_emot0,
			emotethink = function(emote, item, mo)
				if item.skin == SKIN
					emote.color = item.player.skincolor
				else
					emote.color = skins[SKIN].prefcolor
				end
			end,
			tpose = true // identifier for emote wheel
		},
		{ // players
			condition = function(item, player)
				return item.player and item.player.bot == player.bot and not item.player.spectator
			end,
			emotestate = S_PLAY_ICON1,
			emotesound = sfx_emot4,
			emotethink = PlayerIcon
		},
		{ // starposts / goalposts
			condition = function(item, player)
				return (item.type == MT_STARPOST
				and item.spawnpoint and item.spawnpoint.angle/360 > player.starpostnum - 1
				or item.type == MT_SIGN
				and not player.exiting)
				and not player.bot
			end,
			emotesprite = SPR_EMOT,
			emoteframe = C, // !
			emotecolor = SKINCOLOR_SUPERGOLD5,
			emotesound = sfx_emot0,
			emotethink = Question
		},
		{ // sneezeables
			condition = function(item, player)
				return (item.flags & MF_ENEMY or item.player and item.player.ctfteam ~= player.ctfteam) and not player.powers[pw_super]
			end,
			emotesprite = SPR_EMOT,
			emoteframe = B,
			emotecolor = SKINCOLOR_SUPERGOLD5,
			emotethink = Sneeze,
			cusval = 1,
			ignoresightcheck = true,
			sneeze = true // identifier for Fluffy Tails
		},
		{ // monitors
			condition = function(item) return item.flags & MF_MONITOR end,
			emotesprite = SPR_EMOT,
			emoteframe = B, // ?
			emotecolor = SKINCOLOR_SUPERGOLD5,
			emotesound = sfx_emot2,
			emotethink = Question,
			cusval = 1,
			question = true // identifier for emote wheel
		},
		{ // emeralds/tokens/emblems/flags
			condition = function(item)
				return item.type >= MT_TOKEN
				and item.type <= MT_FLINGEMERALD
			end,
			emotesprite = SPR_EMOT,
			emoteframe = B, // ?
			emotecolor = SKINCOLOR_SUPERGOLD5,
			emotesound = sfx_emot2,
			emotethink = Question,
			cusval = 1
		},
		{ // Eggman statues
			condition = function(item) return item.type == MT_EGGSTATUE or item.type == MT_EGGSTATUE2 end,
			emotesprite = SPR_EMOT,
			emoteframe = J, // eggman face
			emotesound = sfx_emot0,
			emotethink = Icon
		},
		{ // pushables/misc
			condition = function(item, player)
				return item.flags & MF_PUSHABLE and not player.powers[pw_carry]
				or item.type == MT_STEAM
			end,
			emotesprite = SPR_EMOT,
			emoteframe = C, // !
			emotecolor = SKINCOLOR_SUPERGOLD5,
			emotesound = sfx_emot0,
			emotethink = Question
		},
		{ // flickies
			condition = function(item) return item.type >= MT_FLICKY_01 and item.type <= MT_SECRETFLICKY_02 end,
			emotesprite = SPR_EMOT,
			emoteframe = D, // <3
			emotecolor = SKINCOLOR_ROSY,
			emotesound = sfx_emot4,
			emotethink = function(emote)
				Bob(emote)
				if leveltime/3 % 7 == 0
					emote.scale = 4*$/3
					emote.z = $ - 8*P_MobjFlip(emote)*emote.scale/3
				end
			end,
			heart = true // identifier for emote wheel
		},
		{ // Mario mode coins
			condition = function(item) return item.type == MT_COIN or item.type == MT_FLINGCOIN end,
			emotesprite = SPR_EMOT,
			emoteframe = M, // $ Fang
			emotecolor = SKINCOLOR_DREAMY,
			emotesound = sfx_mario4,
			emotethink = Cash
		},
		{ // bots
			condition = function(item, player)
				if item.player and item.player.bot ~= player.bot and not item.player.spectator
					local neck = item.neck
					if neck
						if neck.target == player.mo
							player.mo.neck.cc_botskip = true
						elseif player.mo.neck.cc_botskip
						and neck.cc_botskip
							return false
						end
					end
					return true
				end
			end,
			emotestate = S_PLAY_ICON1,
			emotesound = sfx_emot4,
			emotethink = PlayerIcon
		},
		// These only exist for the sake of emote wheel, hence condition always returning false
		{ // clap
			condition = function() return false end,
			emotesprite = SPR_EMOT,
			emoteframe = G,
			emotethink = function(emote)
				if emote.movedir ~= emote.lastlook
					emote.threshold = 2*TICRATE/3
				end
				
				if emote.threshold
					emote.threshold = $ - 1
					if emote.threshold < CLAPTIME
						if emote.threshold == CLAPTIME - 1
							S_StartSound(mo, sfx_emot5)
						end
						emote.frame = $ + 1
					end
				end
			end,
			clap = true
		},
		{ // tailsthink
			condition = function() return false end,
			emotesprite = SPR_EMOT,
			emoteframe = L,
			emotecolor = SKINCOLOR_COPPER,
			emotesound = sfx_emot0,
			emotethink = function(emote, item, mo)
				Bob(emote)
				if mo.skin == "fluffy"
					emote.color = mo.player.skincolor
				else
					emote.color = skins["fluffy"].prefcolor
				end
			end,
			thinking = true
		},
		{ // high-five
			condition = function() return false end,
			emotesprite = SPR_EMOT,
			emoteframe = N,
			emotesound = sfx_emot4,
			emotethink = function(emote)
				emote.threshold = ($ + 1) % 40
				if emote.threshold < 20
					local n = emote.threshold % 10
					emote.scale = $ + (sin(n*ANG2*9) >> 1)
				end
			end,
			highfive = true
		}
	}
}

// Emote wheel table setup
for i, item in ipairs(CC_WHEEL_EMOTES)
	local found = false
	for j, trackable in ipairs(CC_NeckSkins[SKIN].trackables)
		if trackable[item.id]
			CC_WHEEL_EMOTES[item.id] = j
			item.id = j
			found = true
			break
		end
	end
	assert(found, "No wheel emote found for identifier "..item.id)
end

if not CC_TrickSkins 
	rawset(_G, "CC_TrickSkins", {})
end
CC_TrickSkins[SKIN] = { // Tricks support
	alwaysenabled = true,
	bouncetricks = true,
	nosupertricks = true,
	{
		rarity = 2*FRACUNIT, // twice as likely to show up
		sprite2 = SPR2_FLIP,
		tics = 2,
		frames = 7,
	},
	{
		sprite2 = SPR2_EDGE,
		rotation = 0,
		tics = 4,
		frames = 4,
	},
	{
		sprite2 = SPR2_GASP,
		tics = 0,
		rotation = 0,
	},
	{
		sprite2 = SPR2_FAN_,
		tics = 0,
		rotation = "random",
	},
	{
		sprite2 = SPR2_RIDE,
		rotation = 1,
		tics = 0,
	},
	{
		sprite2 = SPR2_SKID,
		rotation = 0,
		tics = 0,
	},
	{
		sprite2 = SPR2_WALK,
		rotation = 0,
		tics = 2,
		frames = 8,
	}
}

local DD_KILLER = 1 << 0
local DD_RACEWIN = 1 << 1

if not CC_DanceSkins
	rawset(_G, "CC_DanceSkins", {})
end
CC_DanceSkins[SKIN] = {
	[1] = { // Fortnite dance, only occurs under certain conditions. Otherwise, is skipped
		condition = function(player)
			local chrispy = player.chrispy
			local pointy = chrispy and player.chrispy.pointy
			if not pointy return end
			
			if chrispy.flags & CF_DEFAULTDANCE return true end
			
			if chrispy.flags & (CF_BOUNCING|CF_CROUCHING)
				return false
			end
			
			if pointy.defaultdance & DD_KILLER
				chrispy.flags = $ | CF_DEFAULTDANCE
				pointy.defaultdance = $ & ~DD_KILLER
				player.mo.state = S_POINTY_DEFAULTDANCE1
				return true
			end
		end
		// has no keys, so dance script will abort if this is selected - but this is what we want!
	},
	[2] = { // Shoot 1
		condition = function(player)
			local chrispy = player.chrispy
			return not (chrispy and chrispy.flags & (CF_BOUNCING|CF_CROUCHING))
		end,
		keys = {
			{
				sprite2 = SPR2_DNCE,
				frame = A,
				tics = 2,
				angle = -ANGLE_90,
				think = function(player) // forwards thrust
					local mo = player.mo
					P_Thrust(mo, (player.cmd.angleturn << 16) - ANGLE_90, 2*mo.scale)
				end
			},
			{
				sprite2 = SPR2_DNCE,
				frame = B,
				tics = 2,
				angle = -ANGLE_90
			},
			{
				sprite2 = SPR2_DNCE,
				frame = C,
				tics = 4,
				angle = -ANGLE_90
			},
			{
				sprite2 = SPR2_DNCE,
				frame = D,
				tics = 3,
				angle = -ANGLE_90,
				think = function(player) // stop
					local mo = player.mo
					mo.momx = player.cmomx
					mo.momy = player.cmomy
				end
			}
		}
	},
	[3] = { // Shoot 2
		condition = function(player)
			local chrispy = player.chrispy
			return not (chrispy and chrispy.flags & (CF_BOUNCING|CF_CROUCHING))
		end,
		keys = {
			{
				sprite2 = SPR2_DNCE,
				frame = E,
				tics = 2,
				angle = -ANGLE_90,
			},
			{
				sprite2 = SPR2_DNCE,
				frame = F,
				tics = 2,
				angle = -ANGLE_90
			},
			{
				sprite2 = SPR2_DNCE,
				frame = G,
				tics = 4,
				angle = -ANGLE_90,
				think = function(player) // stop
					local mo = player.mo
					mo.momx = player.cmomx
					mo.momy = player.cmomy
				end
			},
			{
				sprite2 = SPR2_DNCE,
				frame = H,
				tics = 3,
				angle = -ANGLE_90,
				think = function(player) // backwards thrust
					local mo = player.mo
					P_Thrust(mo, (player.cmd.angleturn << 16) - ANGLE_90, -2*mo.scale)
				end
			}
		}
	}
}

CC_Footsteps[SKIN] = {
	[SPR2_BODY] = {
		flags = FSF_NOTRANSITION,
		[A] = sfx_rstep1
	},
	[SPR2_WAIT] = {
		flags = FSF_NOTRANSITION,
		[A] = sfx_rstep1
	},
	[SPR2_WALK] = {
		[C] = sfx_rstep1,
		[G] = sfx_rstep2
	},
	[SPR2_RUN] = {
		[B] = sfx_rstep2, 
		[D] = sfx_rstep2
	},
	[SPR2_DASH] = {
		flags = FSF_MOVEMENTONLY,
		repeatsound = sfx_squeal
	},
	[SPR2_ROLL] = {
		flags = FSF_MOVEMENTONLY,
		repeatsound = sfx_marble
	},
	[SPR2_SPIN] = {
		flags = FSF_MOVEMENTONLY,
		repeatsound = sfx_marble
	},
	[SPR2_POSE] = {
		flags = FSF_NOTRANSITION|FSF_MOVEMENTONLY,
		[A] = sfx_rstep1*FRACUNIT + sfx_rstep2,
		repeatsound = sfx_squeal
	},
	[SPR2_CNT1] = {
		flags = FSF_NOTRANSITION|FSF_MOVEMENTONLY,
		[A] = sfx_rstep1,
		repeatsound = sfx_rskid2
	},
	[FF_SPR2SUPER|SPR2_WALK] = {},
	[FF_SPR2SUPER|SPR2_RUN] = {},
	[FF_SPR2SUPER|SPR2_WAIT] = {}
}

local FLAGS = {
	CC_RegisterFlag("CF_PEELOUT"),
	CC_RegisterFlag("CF_PEELCARRY"),
	CC_RegisterFlag("CF_AFTERIMAGES"),
	CC_RegisterFlag("CF_BOUNCING"),
	CC_RegisterFlag("CF_FLIPPED"),
	CC_RegisterFlag("CF_DEFAULTDANCE"),
	CC_RegisterFlag("CF_SUPERPOINTY"),
	CC_RegisterFlag("CF_AUTOBRAKE")
}
rawset(_G, "CF_POINTYMASK", FLAGS[#FLAGS]|(FLAGS[#FLAGS] - 1) - (FLAGS[1] - 1))
FLAGS = nil

rawset(_G, "CC_SpawnDustRing", function(mo, speed, thrust, alwaysabove)
	local momz = mo.momz
	if alwaysabove
		momz = -P_MobjFlip(mo)*abs($)
	end
	if abs(momz) < mo.scale
		momz = $ < 0 and -mo.scale or mo.scale
	end
		
	local forwardangle = R_PointToAngle2(0, 0, mo.momx, mo.momy)
	local sideangle = forwardangle + ANGLE_90
	local vangle = R_PointToAngle2(0, 0, FixedHypot(mo.momx, mo.momy), momz)
	
	local cosine = cos(vangle)
	local sine = sin(vangle)
	
	local radius = FixedDiv(mo.height, mo.scale) >> 1
	local xspawn = -FixedMul(FixedMul(cos(forwardangle), cosine), radius)
	local yspawn = -FixedMul(FixedMul(sin(forwardangle), cosine), radius)
	local zspawn = -FixedMul(sine, radius)
	
	local hthrust = 0
	local vthrust = 0
	if thrust
		hthrust = FixedMul(thrust, cosine)
		vthrust = FixedMul(thrust, sine)
	end
	
	cosine = FixedMul($, speed)
	sine = FixedMul($, speed)
	
	for i = 1, 16
		local dust = P_SpawnMobjFromMobj(mo, xspawn, yspawn, radius, mo.eflags & (MFE_UNDERWATER | MFE_TOUCHWATER) and MT_MEDIUMBUBBLE or MT_DUST)
		
		local a = i*ANGLE_22h + forwardangle
		
		local forwardthrust = FixedMul(cos(a), sine)
		local sidethrust = FixedMul(sin(a), speed)
		local zthrust = FixedMul(cosine, cos(a))
		
		dust.z = $ + zspawn
		
		P_Thrust(dust, forwardangle, forwardthrust - hthrust)
		P_Thrust(dust, sideangle, sidethrust)
		dust.momz = -zthrust - vthrust
		
		if dust.type == MT_DUST then dust.state = S_CHRISPY_DUST2 end
		dust.fuse = TICRATE/4
		dust.destscale = mo.scale/8
	end
end)

// Default Dance functions
local DEFAULTDANCE_MUSIC = {
	["TECHNO HILL"] = "THZDD",
	["DEEP SEA"] = "DSZDD",
	["MEADOW MATCH"] = "MMZDD"
}
local dancingmobj

local function StartDefaultDanceMusic(mo)
	local dplayers = {displayplayer, secondarydisplayplayer}
	for i, player in ipairs(dplayers)
		local pmo = player.realmo
		local dist
		
		if not pmo continue end // shouldn't happen? but not taking chances
		
		dist = FixedHypot(FixedHypot(mo.x - pmo.x, mo.y - pmo.y), mo.z + (mo.height >> 1) - (pmo.z + (pmo.height >> 1)))
		if dist < 512*mo.scale
			S_ChangeMusic(0, false) // force a restart to interrupt other default dances!
			S_ChangeMusic(DEFAULTDANCE_MUSIC[mapheaderinfo[gamemap] and mapheaderinfo[gamemap].lvlttl:upper()] or "GFZDD", false)
			dancingmobj = mo
			break
		end
	end
end

local function RunDefaultDanceMusic()
	if dancingmobj
		if not (dancingmobj.valid and dancingmobj.state >= S_POINTY_DEFAULTDANCE1 and dancingmobj.state <= S_POINTY_DEFAULTDANCE2)
			local dplayers = {consoleplayer, secondarydisplayplayer}
			for _, player in ipairs(dplayers)
				P_RestoreMusic(player)
			end
			dancingmobj = nil
		end
	end
end

// Main loop
local function RunPointyThinkframe()
	for player in players.iterate
		local mo = player.mo
		if not mo
		or not player.chrispy
			continue
		end
		
		local chrispy = player.chrispy
		local pointy = chrispy.pointy
		
		if mo.skin ~= SKIN
			if pointy
				chrispy.pointy = nil
				chrispy.flags = $ & ~CF_POINTYMASK
			end
			continue
		end
		
		if player.powers[pw_carry] == CR_NIGHTSMODE
			continue
		end
		
		// set up custom variables
		
		if not chrispy.pointy
			pointy = {}
			chrispy.pointy = pointy
			pointy.mo = mo
			pointy.player = player
			pointy.prevpos = {mo.x, mo.y, mo.z}
			pointy.prevprevpos = pointy.prevpos
			pointy.prevpflags = player.pflags
			pointy.bouncemomentum = {mo.momx, mo.momy}
			pointy.fliptime = 0
			pointy.defaultdance = 0
			pointy.grabtime = 0
			pointy.imouttahere = 0
			chrispy.revheight = -14
			chrispy.superanims = true
			
			player.thrustfactor = skins[mo.skin].thrustfactor
		end
		
		// locals
		
		local flip = P_MobjFlip(mo)
		local grounded = P_IsObjectOnGround(mo)
		local looking = mo.state == S_PLAY_LOOK
		local crouchbutton = player.cc_buttonmap and player.cc_buttonmap.crouch or CC_DEFAULTBUTTONS.crouch
		
		// grounded checks
		
		if grounded
		and not pointy.grounded
			chrispy.flags = $ & ~CF_FLIPPED
			if chrispy.flags & CF_BOUNCING
				pointy.bouncemomentum = {mo.momx, mo.momy, mo.momz}
				mo.state = S_POINTY_BOUNCE_LANDING
				S_StartSound(mo, sfx_rbound)
				if player.powers[pw_super]
					P_Earthquake(mo, mo, 256*FRACUNIT)
					S_StartSound(mo, sfx_pstop)
				end
			end
		end
		
		// Default dance
		
		if chrispy.flags & CF_DEFAULTDANCE
			if mo.state >= S_POINTY_DEFAULTDANCE1
			and mo.state <= S_POINTY_DEFAULTDANCE2
				if mo.frame & FF_FRAMEMASK == B
				and mo.tics == 3
					StartDefaultDanceMusic(mo)
				end
				player.pflags = $ | PF_FULLSTASIS
				player.drawangle = (player.cmd.angleturn << 16) + ANGLE_135
			else
				chrispy.flags = $ & ~CF_DEFAULTDANCE
			end
		end
		
		// Bounce
		
		if srb2p
			// nah
		elseif chrispy.buttons[crouchbutton] == 1
		and pointy.prevpflags & PF_JUMPED
		and player.pflags & (PF_JUMPED|PF_NOJUMPDAMAGE|PF_SLIDING|PF_GLIDING|PF_BOUNCING) == PF_JUMPED
		and not (chrispy.flags & CF_BOUNCING)
		and not (mo.state >= S_PLAY_SUPER_TRANS1 and mo.state <= S_PLAY_SUPER_TRANS6)
			if not (player.cmd.forwardmove or player.cmd.sidemove)
				mo.momx = 0
				mo.momy = 0
			end
			mo.state = S_POINTY_BOUNCE1
			mo.frame = 0
			for i = 1, 4
				mo.state = $
			end
			player.pflags = $ & ~(PF_SPINNING|PF_NOJUMPDAMAGE|PF_STARTJUMP) | PF_THOKKED
			chrispy.flags = $ & ~(CF_DROPDASHCHARGE|CF_FLIPPED) | CF_BOUNCING
			if mo.eflags & MFE_GOOWATER
				mo.momz = $ - flip*FixedMul(skins[SKIN].actionspd/3, mo.scale)
				pointy.goopbounce = 12
			else
				mo.momz = flip*(min(flip*$, 0) - FixedMul(skins[SKIN].actionspd, mo.scale))
				pointy.goopbounce = 0
			end
			CC_SpawnDustRing(mo, 6*mo.scale, 2*mo.scale, true)
			S_StartSound(mo, sfx_rbnce)
		end
		
		if chrispy.flags & CF_BOUNCING
			// exit conditions
			if P_PlayerInPain(player)
			or player.powers[pw_carry] ~= CR_NONE
			or player.playerstate ~= PST_LIVE
			or player.pflags & (PF_SPINNING|PF_SLIDING|PF_GLIDING|PF_BOUNCING)
			or player.panim ~= PA_ABILITY and mo.state ~= S_POINTY_BOUNCE_LANDING and not (mo.eflags & MFE_SPRUNG)
				chrispy.flags = $ & ~CF_BOUNCING
				if mo.state == S_POINTY_BOUNCE1
				or mo.state == S_POINTY_BOUNCE2
					mo.state = S_PLAY_JUMP
				end
			elseif pointy.goopbounce == 1
				chrispy.flags = $ & ~CF_BOUNCING
				player.pflags = $ & ~PF_THOKKED
				for i = 1, 12
					mo.state = S_PLAY_JUMP
				end
			elseif flip*mo.momz > 0
				if mo.eflags & MFE_SPRUNG
					chrispy.flags = $ & ~CF_BOUNCING
					S_StartSound(mo, sfx_sprong)
					mo.momz = 6*$/5 // yes, I know this doesn't account for the type of spring hit. no, I don't care. maybe I should add a spring hook to SRB2.
					mo.momx = 4*$/3
					mo.momy = 4*$/3
					mo.state = S_POINTY_FLIP
				elseif not pointy.goopbounce
					chrispy.flags = $ & ~CF_BOUNCING
					player.pflags = $ & ~PF_THOKKED
					for i = 1, 12
						mo.state = S_PLAY_JUMP
					end
					if abs(mo.momz) > FixedMul(skins[SKIN].actionspd, mo.scale)/3
					and not pointy.goopbounce
						mo.momz = 3*$/4
						S_StartSound(mo, sfx_rbound)
					end
				end
			end
			
			if pointy.goopbounce
				pointy.goopbounce = $ - 1
			end
			
			if mo.state == S_POINTY_REBOUND
			and mo.tics == 1
				chrispy.flags = $ & ~CF_BOUNCING
				mo.z = $ + flip
				player.pflags = $ | PF_JUMPED
				if chrispy.buttons[BT_JUMP] > 0
				and chrispy.buttons[BT_JUMP] <= states[S_POINTY_BOUNCE_LANDING].tics + states[mo.state].tics
					player.pflags = $ | PF_JUMPDOWN
					for i = 1, 12
						mo.state = S_PLAY_JUMP
					end
				else
					player.pflags = $ | PF_NOJUMPDAMAGE | PF_THOKKED
				end
				mo.momx, mo.momy = unpack(pointy.bouncemomentum)
				mo.momz = flip*max(flip*pointy.bouncemomentum[3], FixedMul(skins[SKIN].actionspd, mo.scale))
				if mo.eflags & MFE_UNDERWATER
					mo.momz = 2 * $ / 3
				end
				if not player.powers[pw_super]
					CC_SpawnDustRing(mo, 6*mo.scale)
				end
			elseif mo.state == S_POINTY_BOUNCE_LANDING
				player.pflags = $ | PF_FULLSTASIS
				mo.momx = 0
				mo.momy = 0
				mo.momz = 0
			end
		end
		
		// Super Pointy controls
		
		if player.powers[pw_super]
			local botfly = player.bot == 1 and players[0].powers[pw_tailsfly]
			local jumpball = (player.pflags & (PF_JUMPED|PF_NOJUMPDAMAGE)) == PF_JUMPED
			local nocontrol = player.pflags & (PF_STASIS|PF_SLIDING)
				or player.powers[pw_nocontrol]
				or player.powers[pw_carry]
				or chrispy.flags & (CF_BOUNCING|CF_EMOTEWHEEL)
				or mo.state == S_CHRISPY_ACT
			
			if botfly and jumpball
				player.pflags = $ & ~PF_JUMPED | PF_THOKKED // enable flying!
				jumpball = 0
			end
			
			// toggle gravity
			if jumpball
			or nocontrol
				mo.flags = $ & ~MF_NOGRAVITY
			else
				if (chrispy.buttons[crouchbutton] == 1)// or player.bot == 1 and not botfly)
				and not grounded
					player.pflags = $ & ~(PF_THOKKED|PF_NOJUMPDAMAGE) | PF_JUMPED | PF_JUMPDOWN
					mo.flags = $ & ~MF_NOGRAVITY
					mo.state = S_PLAY_JUMP
					jumpball = 1
				else
					if chrispy.flags & CF_CROUCHING
						player.pflags = $ & ~PF_JUMPSTASIS
					elseif player.bot ~= 1 or botfly
						player.pflags = $ | PF_JUMPSTASIS
					end
					if grounded or player.pflags & PF_SPINNING or mo.state == S_CHRISPY_DANCE
						mo.flags = $ & ~MF_NOGRAVITY
					else
						mo.flags = $ | MF_NOGRAVITY
					end
				end
			end
			
			// tighter controls in midair to account for lack of friction
			if mo.flags & MF_NOGRAVITY
				player.thrustfactor = 2*skins[mo.skin].thrustfactor
			else
				player.thrustfactor = skins[mo.skin].thrustfactor
			end
			
			// rise/fall controls
			if not nocontrol
			and not jumpball
				if chrispy.buttons[BT_JUMP] // rise with jump
				and not (grounded and chrispy.buttons[BT_USE])
				and not (chrispy.flags & CF_CROUCHING)
					player.pflags = $ & ~PF_SPINNING
					if mo.state == S_CHRISPY_DANCE
						mo.state = S_PLAY_WALK
					end
					mo.momz = $ + P_MobjFlip(mo)*mo.scale
				end
				
				if chrispy.buttons[BT_USE] and not grounded // sink with spin
				and not (player.pflags & PF_SPINNING)
					if mo.state == S_CHRISPY_DANCE
						mo.state = S_PLAY_WALK
					end
					mo.momz = $ - P_MobjFlip(mo)*mo.scale
				end
			end
			
			// reduce vertical speed to 0 when not manually rising or falling
			if mo.flags & MF_NOGRAVITY
			and mo.momz
			and ((chrispy.buttons[BT_JUMP] and true) == (chrispy.buttons[BT_USE] and true))
				mo.momz = ($ - ($/abs($)) * min(abs(P_GetMobjGravity(mo)), abs($))) - flip
			end
			
			// do not allow landing into a spin
			if not grounded
			and not (player.pflags & (PF_JUMPED|PF_SPINNING))
				player.pflags = $ | PF_THOKKED
			end
			
			// fake autobrake stuff
			local dirinput = player.cmd.forwardmove or player.cmd.sidemove
			
			if player.pflags & PF_APPLYAUTOBRAKE
			or mo.eflags & MFE_SPRUNG
			or grounded
			or not (mo.flags & MF_NOGRAVITY)
				chrispy.flags = $ & ~CF_AUTOBRAKE
			elseif dirinput
			or player.cmd.buttons & (BT_JUMP|BT_USE)
				chrispy.flags = $ | CF_AUTOBRAKE
			end
			
			if chrispy.flags & CF_AUTOBRAKE
			and not dirinput
			and not (chrispy.flags & CF_DABBING)
			and player.speed
				CC_Autobrake(player)
			end
			
			chrispy.flags = $ | CF_SUPERPOINTY
		elseif chrispy.flags & CF_SUPERPOINTY
			chrispy.flags = $ & ~(CF_SUPERPOINTY|CF_AUTOBRAKE)
			player.pflags = $ | PF_JUMPDOWN
			mo.flags = $ & ~MF_NOGRAVITY
			
			player.thrustfactor = skins[mo.skin].thrustfactor
		end
		
		// Momentum flip
		
		if mo.state == S_POINTY_FLIP
		and chrispy.flags & CF_FLIPPED
			if pointy.fliptime
			and (player.cmd.forwardmove or player.cmd.sidemove)
				local momangle = R_PointToAngle2(0, 0, mo.momx, mo.momy)
				local inputangle = player.cmd.angleturn << 16 + R_PointToAngle2(0, 0, player.cmd.forwardmove << 16, -player.cmd.sidemove << 16)
				local anglediff = inputangle - momangle
				if abs(anglediff) < ANGLE_112h
					local newangle = momangle + (anglediff)/pointy.fliptime
					P_InstaThrust(mo, newangle, FixedHypot(mo.momx, mo.momy))
				end
			end
			pointy.fliptime = $ + 1
		elseif pointy.fliptime
			pointy.fliptime = 0
		end
		
		// Spindash visual
		
		if player.pflags & (PF_STARTDASH|PF_JUMPED) == PF_STARTDASH
			if player.dashspeed >= player.maxdash
				if mo.state ~= S_PLAY_SPINDASH
					mo.state = S_PLAY_SPINDASH
				end
			elseif player.dashspeed >= player.mindash + 2*(player.maxdash - player.mindash)/3
				if mo.state ~= S_CHRISPY_DROPDASH
					mo.state = S_CHRISPY_DROPDASH
				end
			else
				if mo.state ~= S_PLAY_ROLL
					mo.state = S_PLAY_ROLL
				end
				if player.dashspeed >= player.mindash + (player.maxdash - player.mindash)/3
					mo.tics = min($, 1)
				end
			end
		end
		
		// Peelout
		
		local peelspeed = twodlevel and 2*PEELOUT_SPEED/3 or PEELOUT_SPEED
		
		if srb2p
			// peelout works in SRB2P, but Fluffy's flight charge doesn't, so let's be consistent
		elseif mo.state == S_CHRISPY_CROUCH
		and not (chrispy.flags & CF_EMOTEWHEEL)
		and chrispy.buttons[BT_JUMP] == 1
		and grounded
		and not player.powers[pw_super]
			mo.momx = player.cmomx
			mo.momy = player.cmomy
			mo.state = S_POINTY_PEELOUT
			S_StartSound(mo, sfx_cdfm11)
			chrispy.flags = $ & ~CF_CROUCHING
			player.height = skins[mo.skin].height
			player.drawangle = mo.angle
			for i = 1, 6
				SpawnChrispyDust(player, FixedMul(FixedDiv(3*peelspeed/2, player.maxdash), mo.scale), mo.scale << 1, -14*mo.scale, true)
			end
		end
		
		if mo.state == S_POINTY_PEELOUT
			if not chrispy.buttons[BT_JUMP]
				P_InstaThrust(mo, mo.angle, FixedMul(peelspeed, mo.scale))
				S_StartSound(mo, sfx_cdfm01)
				mo.state = S_PLAY_RUN
				//pointy.grabtime = PEELOUT_GRAB_TIME
			else
				if player.pflags & (PF_DIRECTIONCHAR|PF_ANALOGMODE) == PF_ANALOGMODE
					mo.angle = player.cmd.angleturn << 16
				end
				player.drawangle = mo.angle
				player.pflags = $ | PF_FULLSTASIS | PF_JUMPDOWN
				for i = 1, 4
					SpawnChrispyDust(player, FixedMul(FixedDiv(peelspeed, player.maxdash), mo.scale), mo.scale >> 2, -14*mo.scale)
				end
			end
		end
		
		local speed = FixedDiv(FixedHypot(mo.momx - player.cmomx, mo.momy - player.cmomy), mo.scale)
		local normalspeed = skins[SKIN].normalspeed
		//local scaledspeed = twodlevel and 3*speed/2 or speed
		//local scalednormalspeed = twodlevel and 3*normalspeed/2 or normalspeed
		
		if speed < normalspeed - FRACUNIT
		or player.powers[pw_super]
			if chrispy.flags & CF_PEELOUT
				player.normalspeed = normalspeed
				player.charflags = $ & ~SF_RUNONWATER
				chrispy.flags = $ & ~(CF_PEELOUT|CF_PEELCARRY)
				//pointy.grabtime = 0
				if mo.state == S_PLAY_DASH
					mo.state = S_PLAY_RUN
				end
			end
		elseif player.powers[pw_sneakers]
			if player.normalspeed > normalspeed
			and chrispy.flags & CF_PEELOUT
				player.normalspeed = normalspeed
			end
		elseif not player.isxmomentum // support for xmom was highly requested on Twitter so
			if speed > player.normalspeed + SPEED_INCREASE_LEEWAY // if above normalspeed, update it with current speed
			or speed < player.normalspeed - SPEED_DECREASE_LEEWAY
				player.normalspeed = min(speed, PEELOUT_SPEED)
			end
			
			if player.normalspeed > normalspeed // slowly decrease normalspeed over time
				player.normalspeed = max(normalspeed, $ - FRACUNIT >> 4)
			end
		end
		
		if (player.normalspeed > normalspeed or player.powers[pw_sneakers])
		and speed >= normalspeed
			if player.panim == PA_RUN
				mo.state = S_PLAY_DASH
			end
			chrispy.flags = $ | CF_PEELOUT
			if not player.isxmomentum // let xmom control runonwater
				if chrispy.flags & CF_BOUNCING
					player.charflags = $ & ~SF_RUNONWATER
				else
					player.charflags = $ | SF_RUNONWATER
				end
			end
		end
		
		if mo.state == S_PLAY_DASH
			mo.tics = min($, 1)
		end
		
		if chrispy.flags & CF_PEELOUT
		and (player.normalspeed > normalspeed + 2*SPEED_INCREASE_LEEWAY
		or player.powers[pw_sneakers] and speed > normalspeed + 2*SPEED_INCREASE_LEEWAY)
			if mo.state == S_PLAY_DASH
			and player.normalspeed > (normalspeed + 2*SPEED_INCREASE_LEEWAY + PEELOUT_SPEED) >> 1
			and grounded
			and not (chrispy.flags & CF_PEELCARRY)
				for i = 1, 4
					SpawnChrispyDust(player, FixedMul(FixedDiv(peelspeed, player.maxdash), mo.scale), mo.scale >> 2, -14*mo.scale)
				end
				//pointy.grabtime = $ - 1
			end
			if chrispy.time & 1
			and not (chrispy.flags & CF_PEELCARRY)
			and player.charflags & SF_RUNONWATER
				CC_QueueGhost(player, pointy.prevpos, mo)
			end
			player.spinitem = 0
			chrispy.flags = $ | CF_AFTERIMAGES
		elseif chrispy.flags & CF_AFTERIMAGES
			player.spinitem = skins[SKIN].spinitem
			chrispy.flags = $ & ~CF_AFTERIMAGES
		end
		
		// randomized teeter durations
		
		if mo.state == S_PLAY_EDGE
		and mo.tics > TEETER_MAX_TICS
			mo.tics = P_RandomRange(TEETER_MIN_TICS, TEETER_MAX_TICS)
		end
		
		// dab / hater
		
		if mo.state == S_CHRISPY_DAB
			player.drawangle = (player.cmd.angleturn << 16) + ANGLE_180 - ANG1
		elseif mo.state == S_CHRISPY_HATER
			player.drawangle = (player.cmd.angleturn << 16)
		end
		
		// tpose death
		
		if mo.state == S_PLAY_WAIT // count up while in wait frames
		and not player.exiting
			pointy.imouttahere = $ + 1
			if pointy.imouttahere >= TPOSEDEATH_TIME
				mo.state = S_POINTY_TPOSE
				pointy.imouttahere = 0
				S_StartSound(mo, sfx_cdpcm3)
			end
		elseif mo.state == S_POINTY_TPOSE
			local n
			pointy.imouttahere = $ + 1
			n = pointy.imouttahere
			
			player.drawangle = (player.cmd.angleturn << 16) + ANGLE_135
			
			if not mo.health // if you're dead, spin the ghosts
				local camang = (splitscreen and (player.cmd.angleturn << 16) + 1 or R_PointToAngle(mo.x, mo.y)) + ANGLE_90
				local dist = n*mo.scale >> 3
				local ang = n*ANGLE_22h
				local hoff = P_ReturnThrustX(mo, ang, dist)
				local zoff = P_ReturnThrustY(mo, ang, dist)
				local xoff = P_ReturnThrustX(mo, camang, hoff)
				local yoff = P_ReturnThrustY(mo, camang, hoff)
				
				for i, ghost in ipairs(pointy.tposers)
					if not ghost.valid then continue end
					ghost.frame = $ & ~FF_TRANSMASK | (FF_TRANS50 + FF_TRANS10*(5*(ghost.threshold - ghost.tics)/ghost.threshold))
					P_TeleportMove(ghost,
						mo.x + xoff,
						mo.y + yoff,
						mo.z + zoff)
					xoff, yoff, zoff = -$1, -$2, -$3
				end
			elseif n > TPOSEDEATH_HOLD // you didn't move. time to DIE
				P_KillMobj(mo)
				for i = 0, 3
					S_StopSoundByID(mo, sfx_altdi1 + i)
				end
				S_StartSound(mo, sfx_s3k6a)
				mo.state = S_POINTY_TPOSE
				mo.flags = $ & ~MF_NOCLIPHEIGHT | MF_NOGRAVITY
				mo.flags2 = $ | MF2_DONTDRAW
				mo.momz = 0
				pointy.imouttahere = 0
				pointy.tposers = {}
				for i = 1, 2
					local ghost = P_SpawnGhostMobj(mo)
					ghost.frame = $ & ~FF_TRANSMASK | FF_TRANS50
					ghost.tics = 6*ghost.fuse
					ghost.threshold = ghost.tics
					ghost.fuse = 0
					table.insert(pointy.tposers, ghost)
				end
			end
		else
			pointy.imouttahere = 0
		end
		
		// super animations
		
		if player.powers[pw_super]
			local velocity = FixedHypot(mo.momx - player.cmomx, mo.momy - player.cmomy)
			local fixedvelocity = FixedDiv(velocity, mo.scale)
			
			if not grounded
			and not player.powers[pw_carry]
			and (player.pflags & (PF_JUMPED|PF_NOJUMPDAMAGE)) ~= PF_JUMPED
			and not (player.pflags & PF_SPINNING)
			and not (chrispy.flags & (CF_BOUNCING|CF_DABBING))
			and player.panim ~= PA_ABILITY
			and mo.state ~= S_PLAY_GASP
			and mo.state ~= S_CHRISPY_DANCE
				local vz = P_MobjFlip(mo)*mo.momz
				vz = abs($) > 4*mo.scale and $ or 0
				
				// don't use any spin frames
				if player.panim == PA_JUMP
				or player.panim == PA_ROLL
				or mo.state == S_PLAY_SPINDASH
					mo.state = S_PLAY_WALK
				end
				
				// spring/fall sprites
				if vz > velocity
					if mo.state ~= S_PLAY_SPRING
						mo.state = S_PLAY_SPRING
					end
				elseif -vz > velocity
					if mo.state ~= S_PLAY_FALL
						mo.state = S_PLAY_FALL
					end
				elseif mo.state == S_PLAY_SPRING
				or mo.state == S_PLAY_FALL
					mo.state = fixedvelocity >= 5*player.runspeed/3 and S_PLAY_RUN or S_PLAY_WALK
				end
			end
			
			// we use float sprites for other things
			if mo.state == S_PLAY_FLOAT_RUN
				mo.state = S_PLAY_RUN
			elseif mo.state == S_CHRISPY_MIDFALL
				mo.state = S_PLAY_WALK
			end
			
			// float at low speeds, walk at high speeds
			if mo.state == S_PLAY_WALK
				if fixedvelocity < CC_NeckSkins[SKIN].superfloatthreshold
					mo.state = S_PLAY_FLOAT
				end
			elseif mo.state == S_PLAY_FLOAT
				if fixedvelocity >= CC_NeckSkins[SKIN].superfloatthreshold
					mo.state = S_PLAY_WALK
				end
			end
			
			// slow the top speed animation
			if mo.state == S_PLAY_RUN
			and chrispy.time & 1
				mo.tics = $ + 1
			end
		end
		
		// kill headtracking objects if no longer looking
		if looking and mo.state ~= S_PLAY_LOOK
			CC_KillNeck(mo.neck)
		end
		
		// resets
		pointy.prevprevpos = pointy.prevpos
		pointy.prevpos = {mo.x, mo.y, mo.z}
		pointy.grounded = grounded
		pointy.prevpflags = player.pflags
	end
end

local function RunPointyCarry()
	for player2 in players.iterate
		local mo2 = player2.mo
		local mo1 = mo2 and mo2.tracer
		local player1 = valid(mo1) and mo1.player
		
		if player2.powers[pw_carry] == CR_GENERIC
		and player1
			if mo1.state == S_PLAY_DASH
			and mo1.skin == SKIN
				local chrispy1 = mo1.chrispy
				local pointy1 = chrispy1 and chrispy1.pointy
				local fm = player2.followmobj
				if player2.cc_peelcarry == 2
					mo2.state = S_PLAY_HUG
					player2.drawangle = player1.drawangle
				else
					mo2.state = S_PLAY_PAIN
					player2.drawangle = player1.drawangle + ANGLE_180
				end
				P_TeleportMove(mo2,
					mo1.x + P_ReturnThrustX(mo1, player1.drawangle, -mo1.radius - mo2.radius),
					mo1.y + P_ReturnThrustY(mo1, player1.drawangle, -mo1.radius - mo2.radius),
					mo1.z + (mo1.eflags & MFE_VERTICALFLIP and mo1.height - mo2.height or 0))
				player2.pflags = $ | PF_STASIS
				mo2.momx = mo1.momx
				mo2.momy = mo1.momy
				
				if valid(fm) and fm.type == MT_TAILSOVERLAY // specific support for Tails, sorry if your custom followitem gets kinda rekt
					P_TeleportMove(fm,
						mo2.x + P_ReturnThrustX(mo2, player2.drawangle, -(mo2.scale >> 4)),
						mo2.y + P_ReturnThrustY(mo2, player2.drawangle, -(mo2.scale >> 4)),
						mo2.z + (mo2.eflags & MFE_VERTICALFLIP and mo2.height - fm.height or 0))
					fm.state = S_TAILSOVERLAY_PAIN
				end
				
				if pointy1 // spawn ghosts
				and chrispy1.time & 1
				and chrispy1.flags & CF_AFTERIMAGES
					CC_QueueGhost(player2, pointy1.prevprevpos, mo1)
				end
				continue
			end
			player2.powers[pw_carry] = CR_NONE
			mo2.tracer = nil
		end
		
		if player2.cc_peelcarry
			if player1
				if player1.chrispy
					player1.chrispy.flags = $ & ~CF_PEELCARRY
				end
				if not player2.powers[pw_carry]
				and not (player2.pflags & PF_JUMPED)
				and not (mo1.eflags & MFE_SPRUNG)
					mo2.momz = mo1.momz
					mo2.z = mo1.z + P_MobjFlip(mo2)
					if player1.skidtime // yeet
						P_SetObjectMomZ(mo2, 12*FRACUNIT, true)
						mo2.momx = 2*mo1.momx
						mo2.momy = 2*mo1.momy
						player2.pflags = $ | P_GetJumpFlags(player2)
						mo2.state = S_PLAY_JUMP
						S_StartSound(mo2, sfx_s3k51)
					elseif player1.pflags & PF_JUMPED // jump boost
						P_SetObjectMomZ(mo2, 8*FRACUNIT, true)
						mo2.momx = mo1.momx
						mo2.momy = mo1.momy
						player2.pflags = $ | P_GetJumpFlags(player2)
						mo2.state = S_PLAY_JUMP
						S_StartSound(mo2, sfx_ngjump)
					else // drop
						P_DoPlayerPain(player2, mo1, mo1)
						player2.drawangle = player1.drawangle + ANGLE_180
						mo2.momx = 3*mo1.momx/4
						mo2.momy = 3*mo1.momy/4
						if player2.cc_peelcarry == 2
							S_StartSound(mo2, sfx_cdpcm7)
						end
					end
				end
			end
			player2.cc_peelcarry = 0
		end
	end
end

addHook("ThinkFrame", do
	RunDefaultDanceMusic()
	RunPointyThinkframe()
	RunPointyCarry()
end)

// Double jump ability

rawset(_G, "CC_MomentumFlip", function(player)
	local mo = player.mo
	local chrispy = player.chrispy
	
	if not chrispy
		return
	end
	
	local in2d = twodlevel or mo.flags2 & MF2_TWOD or mo.currentaxis
	
	player.pflags = $ & ~PF_JUMPED // remove the jumping flag so the game recognizes him as not already jumping
	P_DoJump(player, false) // Do a second jump!
	player.pflags = $ & ~PF_SPINNING | PF_NOJUMPDAMAGE | PF_THOKKED // mark Pointy as vulnerable again
	mo.state = S_POINTY_FLIP
	if not in2d
		chrispy.flags = $ | CF_FLIPPED
	end
	
	if player.cmd.forwardmove or player.cmd.sidemove // momentum reshift
		if in2d
			P_InstaThrust(mo, mo.angle, FixedHypot(mo.momx, mo.momy))
		else
			P_InstaThrust(mo, player.cmd.angleturn << 16 + R_PointToAngle2(0, 0, player.cmd.forwardmove*FRACUNIT, -player.cmd.sidemove*FRACUNIT), FixedHypot(mo.momx, mo.momy)) // hi Rumia
		end
	end
	
	S_StartSound(mo, sfx_dbjump)
	
	CC_SpawnDustRing(mo, 6*mo.scale, 2*mo.scale)
end)

addHook("AbilitySpecial", function(player)
	local mo = player.mo
	
	if player.charability ~= skins[SKIN].ability
	or not mo
	or mo.skin ~= SKIN
		return
	end
	
	if player.powers[pw_super]
		player.pflags = $ & ~PF_JUMPED | PF_THOKKED // enable flying!
		return true
	elseif player.pflags & PF_THOKKED
		return
	end
	
	CC_MomentumFlip(player)
	
	return true // don't activate the default ability lol
end)

// allow damage during grounded bounce, and to monitors while super

addHook("PlayerCanDamage", function(player)
	local chrispy = player.chrispy
	if chrispy and chrispy.flags & CF_BOUNCING
		return true
	end
	if player.powers[pw_super] and player.mo.skin == SKIN
		return true
	end
end)

// Pointy's spinball overlay

local function PointyOverball(player, ball, mo)
	local chrispy = player.chrispy
	mo = $ or player.mo
	
	if not (mo and chrispy and ball.valid)
		return
	end
	
	local info = ball.info
	local frame = mo.frame & FF_FRAMEMASK
	
	ball.scale = mo.scale
	ball.color = mo.color
	
	if mo.sprite2 == SPR2_ROLL
	and mo.state ~= S_CHRISPY_ACT
		// if spin flicker is disabled, skip flicker frames
		if player.cc_flickerenabled == false
			while (mo.frame & FF_FRAMEMASK) % 3
				mo.state = $
			end
			return
		end
		
		// show the translucency effect every 3rd frame
		if frame % 3 == 1
			ball.skin = mo.skin
			ball.sprite = mo.sprite
			ball.sprite2 = SPR2_TWIN
			ball.frame = (2*(frame/3) + 1) | min(TR_TRANS90, (mo.frame & FF_TRANSMASK) + TR_TRANS50)
			return
		end
	end
	
	// during a spindash charge, "fade" into the spindash sprites during the last 2 frames
	if player.pflags & PF_STARTDASH
	and player.dashspeed >= player.maxdash - 2*FRACUNIT
	and player.dashspeed < player.maxdash
	and player.cc_flickerenabled ~= false
		ball.state = S_PLAY_SPINDASH
		ball.frame = $ | min(TR_TRANS90, (mo.frame & FF_TRANSMASK) + TR_TRANS50)
		return
	end
	
	// not using the effect? then fuck off
	if ball.sprite
		ball.state = info.spawnstate
	end
end

rawset(_G, "CC_PointyOverball", PointyOverball)

addHook("FollowMobj", PointyOverball, MT_POINTY_OVERBALL)

// Default dance trigger

addHook("MobjDeath", function(mo, inflictor, source)
	local pointy = valid(source) and source.skin == SKIN and source.chrispy and source.chrispy.pointy
	
	if not pointy
	or not (mo.valid and (mo.player or mo.flags & MF_BOSS))
		return
	end
	
	pointy.defaultdance = $ | DD_KILLER
end)

// Peelout carry

addHook("MobjMoveCollide", function(mo1, mo2)
	if not mo1.valid and mo2.valid
		return
	end
	
	local player1 = mo1.player
	local player2 = mo2.player
	local chrispy1 = mo1.chrispy
	//local pointy1 = chrispy1 and chrispy1.pointy
	
	if not (player1 and player2)
	or not chrispy1
	or chrispy1.flags & (CF_PEELOUT|CF_PEELCARRY) ~= CF_PEELOUT
	or mo1.z > mo2.z + mo2.height
	or mo2.z > mo1.z + mo1.height
	or player2.powers[pw_carry]
	or player2.playerstate ~= PST_LIVE
	or player2.spectator
	or P_PlayerInPain(player2)
	or mo1.state ~= S_PLAY_DASH
	or mo1.skin ~= SKIN
	or not P_IsObjectOnGround(mo1)
	or player1.normalspeed <= (skins[SKIN].normalspeed + 2*SPEED_INCREASE_LEEWAY + PEELOUT_SPEED) >> 1
		return
	end
	
	if mo2.skin == "amy"
		player2.cc_peelcarry = 2
		S_StartSound(mo2, sfx_cdpcm6)
	else
		player2.cc_peelcarry = 1
	end
	player2.powers[pw_carry] = CR_GENERIC
	chrispy1.flags = $ | CF_PEELCARRY
	mo2.tracer = mo1
	S_StartSound(mo1, sfx_s3k4a)
end, MT_PLAYER)

// carried players shouldn't collide with anything

local function NoTouchy(mo, spring)
	local player = mo.valid and mo.player
	
	if player and player.cc_peelcarry
		return false
	end
end

addHook("MobjMoveCollide", NoTouchy, MT_PLAYER)
addHook("MobjCollide", NoTouchy, MT_PLAYER)