/*
ChrispyChars configuration
	by Lach

beta
*/

// To prevent duplicate freeslots
local function SafeFreeslot(...)
	for _, item in ipairs({...})
		if rawget(_G, item) == nil
			freeslot(item)
		end
	end
end

// hug state, lifted from Hugs by Tatsuru https://mb.srb2.org/showthread.php?t=45828

SafeFreeslot("S_PLAY_HUG", "SPR_HUG_")

states[S_PLAY_HUG] = {
	sprite = SPR_HUG_,
	frame = A,
	tics = -1
}

// Sounds

SafeFreeslot("sfx_null", "sfx_ddash", "sfx_dbjump", "sfx_rbnce", "sfx_rbound",
	"sfx_rskid2", "sfx_dab", "sfx_sneeze", "sfx_rscrch", "sfx_trckhs",
	"sfx_rput", "sfx_rstep1", "sfx_rstep2", "sfx_marble", "sfx_squeal", "sfx_ftdash",
	"sfx_rmenu1", "sfx_rmenua", "sfx_rmenub")

for i = 0, 6
	SafeFreeslot("sfx_emot"..i)
	sfxinfo[sfx_emot0 + i].caption = "/"
end

sfxinfo[sfx_cdfm01].caption = "Peelout"
sfxinfo[sfx_cdfm11].caption = "Peeling"
sfxinfo[sfx_cdfm40].caption = "/"
sfxinfo[sfx_s236].caption = "/"
sfxinfo[sfx_s227].caption = "Twinkle"
sfxinfo[sfx_s25d].caption = "Tink"

sfxinfo[sfx_dab].caption = "/"
sfxinfo[sfx_ddash].caption = "Drop dash"
sfxinfo[sfx_dbjump].caption = "Flip"
sfxinfo[sfx_rbnce].caption = "Bounce"
sfxinfo[sfx_rbound].caption = "Rebound"
sfxinfo[sfx_rskid2].caption = "/"
sfxinfo[sfx_rstep1].caption = "/"
sfxinfo[sfx_rstep2].caption = "/"
sfxinfo[sfx_squeal].caption = "/"
sfxinfo[sfx_marble].caption = "/"
sfxinfo[sfx_ftdash].caption = "/"
sfxinfo[sfx_rput].caption = "/"
sfxinfo[sfx_sneeze].caption = "Sneeze"
sfxinfo[sfx_rmenu1].caption = "Swap"
sfxinfo[sfx_rmenua].caption = "Accept"
sfxinfo[sfx_rmenub].caption = "Back"
sfxinfo[sfx_rscrch].caption = "Record scratch"
sfxinfo[sfx_trckhs].caption = "/"

// Colors

SafeFreeslot("SKINCOLOR_CHRISP", "SKINCOLOR_DREAMY", "SKINCOLOR_TAFFY", "SKINCOLOR_CARDINAL", "SKINCOLOR_BEEG")

local reps
local colors
local hypercolors = {}

for i = SKINCOLOR_SUPERSILVER1, SKINCOLOR_SUPERPURPLE1, 5
	local ramp1 = skincolors[i].ramp
	local ramp5 = skincolors[i + 4].ramp
	local name = skincolors[i].name
	name = $:sub(select(2, $:find("Super ")) + 1)
	name = $:sub(1, $:find(" ") - 1)
	
	table.insert(hypercolors,
		{
			name = name,
			chatcolor = skincolors[1 + 4].chatcolor,
			ramp1[9], ramp1[9], ramp1[11], ramp1[13], ramp1[15], ramp5[12], ramp5[14]
		}
	)
end

local function SetupRamp(...)
	local ramp = {}
	for i, index in ipairs({...})
		local n = reps[i]
		while n
			table.insert(ramp, index)
			n = $ - 1
		end
	end
	return ramp
end

//greens = { 96,  97,  98,  99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111}

reps = {2, 2, 2, 3, 3, 2, 2}
colors = {
	SetupRamp(0, 0, 80, 83, 84, 86, 58),
	SetupRamp(0, 80, 82, 73, 85, 87, 60),
	SetupRamp(80, 81, 72, 84, 86, 245, 70),
	SetupRamp(81, 82, 72, 64, 66, 69, 71),
	SetupRamp(81, 83, 73, 65, 67, 70, 47)
}
for i = 1, 5
	SafeFreeslot("SKINCOLOR_SUPERLIGHTBULB"..i)
	skincolors[SKINCOLOR_SUPERLIGHTBULB1 + i - 1] = {
		name = "Super Lightbulb "..i,
		ramp = colors[i],
		invcolor = SKINCOLOR_SAPPHIRE,
		invshade = 7,
		chatcolor = V_YELLOWMAP,
		accessible = false
	}
end

for i = 1, #hypercolors
	local color = hypercolors[i]
	local name = color.name
	local opposite = skincolors[FIRSTSUPERCOLOR + 5*(i - 1)]
	SafeFreeslot("SKINCOLOR_HYPERPOINTY"..name:upper())
	skincolors[SKINCOLOR_HYPERPOINTYSILVER + i - 1] = {
		name = "Hyper Pointy "..name,
		ramp = SetupRamp(unpack(color)),
		invcolor = opposite.invcolor,
		invshade = opposite.invshade,
		chatcolor = color.chatcolor,
		accessible = false
	}
end

reps = {2, 2, 1, 2, 2, 2, 5}
colors = {
	SetupRamp(0, 0, 0, 160, 160, 171, 174),
	SetupRamp(0, 0, 80, 4, 160, 163, 166),
	SetupRamp(0, 80, 208, 49, 51, 67, 69),
	SetupRamp(80, 208, 49, 51, 67, 68, 70),
	SetupRamp(82, 216, 50, 52, 67, 69, 71)
}
for i = 1, 5
	SafeFreeslot("SKINCOLOR_SUPERSTREETLAMP"..i)
	skincolors[SKINCOLOR_SUPERSTREETLAMP1 + i - 1] = {
		name = "Super Streetlamp "..i,
		ramp = colors[i],
		invcolor = SKINCOLOR_SAPPHIRE,
		invshade = 7,
		chatcolor = V_ORANGEMAP,
		accessible = false
	}
end

for i = 1, #hypercolors
	local color = hypercolors[i]
	local name = color.name
	local opposite = skincolors[FIRSTSUPERCOLOR + 5*(i - 1)]
	SafeFreeslot("SKINCOLOR_HYPERFLUFFY"..name:upper())
	skincolors[SKINCOLOR_HYPERFLUFFYSILVER + i - 1] = {
		name = "Hyper Fluffy "..name,
		ramp = SetupRamp(unpack(color)),
		invcolor = opposite.invcolor,
		invshade = opposite.invshade,
		chatcolor = color.chatcolor,
		accessible = false
	}
end
reps = nil
colors = nil
hypercolors = nil

rawset(_G, "CC_TRANSCOLOR_SLOTS", 4)
for i = 1, CC_TRANSCOLOR_SLOTS // four slots for super transformations, I don't really want to freeslot a bunch of resources for this but 4 sounds pretty reasonable to me
	SafeFreeslot("SKINCOLOR_CHRISPYTRANS"..i)
	skincolors[SKINCOLOR_CHRISPYTRANS1 + i - 1] = {
		name = "Chrispy Transformation (slot "..i..")",
		ramp = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
		invcolor = SKINCOLOR_CHRISP,
		invshade = 7,
		chatcolor = 0,
		accessible = false
	}
end

skincolors[SKINCOLOR_CHRISP] = { // all-black color for fire death & menu shadows
	name = "Chrisp",
	ramp = {31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31, 31},
	invcolor = SKINCOLOR_FLAME,
	invshade = 7,
	chatcolor = V_GRAYMAP,
	accessible = false
}

skincolors[SKINCOLOR_DREAMY] = { // NiGHTS
	name = "Dreamy",
	ramp = {178, 180, 181, 193, 194, 194, 195, 195, 196, 196, 197, 197, 198, 198, 199, 253},
	invcolor = SKINCOLOR_SUPERLIGHTBULB3,
	invshade = 7,
	chatcolor = V_PURPLEMAP,
	accessible = true
}

skincolors[SKINCOLOR_TAFFY] = { // pink bird
	name = "Taffy",
//greens = { 96,  97,  98,  99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111}
  //ramp = {176, 177, 178, 179, 180, 180, 181, 182, 195, 196, 196, 185, 185, 185, 186, 186}, (Pre-Lachification)
	ramp = {176, 177, 178, 179, 180, 181, 182, 193, 194, 195, 196, 185, 186, 186, 187, 187},
	invcolor = SKINCOLOR_BEEG,
	invshade = 1,
	chatcolor = V_MAGENTAMAP,
	accessible = true
}

skincolors[SKINCOLOR_CARDINAL] = { // red bird
	name = "Cardinal",
//greens = { 96,  97,  98,  99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111}
  //ramp = { 32,  33,  34,  34,  35,  35,  37,  38,  61,  61,  62,  63,  70,  71,  71,  71},
	ramp = { 49,  51,  32,  33,  34,  35,  38,  39,  60,  61,  62,  63,  44,  71, 198, 199},
	invcolor = SKINCOLOR_SUPERLIGHTBULB5,
	invshade = 7,
	chatcolor = V_REDMAP,
	accessible = true
}

skincolors[SKINCOLOR_BEEG] = { // BEEG BIRD
	name = "Beeg",
//greens = { 96,  97,  98,  99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111}
  //ramp = {112, 113,  32, 114,  34, 115,  38, 117, 119,  61,  62,  63,  44,  71, 198, 199},
	ramp = { 73, 188, 112, 113, 114, 115, 116, 116, 117, 117, 118, 118, 119, 119, 253, 139},
	invcolor = SKINCOLOR_TAFFY,
	invshade = 6,
	chatcolor = V_GREENMAP,
	accessible = true
}

// Skin sprites

SafeFreeslot("SPR2_FLIP", "SPR2_FAN_", "SPR2_POSE", "SPR2_CARY", "SPR2_SNEZ", "SPR2_DNCE",
	"SPR2_FROZ", "SPR2_PDED", "SPR2_TPOS", "SPR2_DFAL", "SPR2_TDAB", "SPR2_HI5_",
	"SPR2_BATL")

spr2defaults[SPR2_FLIP] = SPR2_SPNG
spr2defaults[SPR2_FAN_] = SPR2_FALL
spr2defaults[SPR2_POSE] = SPR2_SPNG
spr2defaults[SPR2_CARY] = SPR2_FLY
spr2defaults[SPR2_SNEZ] = SPR2_HEDF
spr2defaults[SPR2_PDED] = SPR2_DEAD
spr2defaults[SPR2_TDAB] = SPR2_POSE
spr2defaults[SPR2_FROZ] = SPR2_DEAD
spr2defaults[SPR2_TPOS] = SPR2_DEAD
spr2defaults[SPR2_DFAL] = SPR2_DNCE
spr2defaults[SPR2_BATL] = SPR2_BATL

// Sprites

SafeFreeslot("SPR_CTHK", "SPR_CHSP", "SPR_RSTR", "SPR_ARM_",
	"SPR_3DFB", "SPR_3DFP", "SPR_3DFR", "SPR_3DFG", "SPR_REMG", "SPR_EMOT", "SPR_NIGH")

local pivot = {}
for i = A, L
	pivot[R_Frame2Char(i)] = {x = 45, y = 45, rotaxis = ROTAXIS_Z}
end
spriteinfo[SPR_NIGH].pivot = pivot

// States

function A_SetPAnim(mo, var1, var2)
	local player = mo.player
	if not player return end
	player.panim = var2
end

function A_ModifyPFlags(mo, var1, var2) // var1 flags are added, var2 flags are removed
	local player = mo.player
	if not player return end
	player.pflags = $ & ~var2 | var1
end

function A_WaterState(mo, var1)
	if mo.eflags & (MFE_UNDERWATER|MFE_TOUCHWATER)
		mo.state = var1
	end
end

function A_Contspeed(mo, var1, var2)
	if mo.skin
		mo.tics = skins[mo.skin].contspeed
	end
	A_SetPAnim(mo, var1, var2)
end

function A_ChrispyFlicky(mo)
	mo.sprite = mo.extravalue2
end

local function A_RandomFrame(mo, var1, var2)
	mo.frame = $ & ~FF_FRAMEMASK | P_RandomRange(var1, var2)
end

SafeFreeslot("S_CHRISPY_SPARKLE", "S_CHRISPY_REVORB", "S_CHRISPY_DUST1", "S_CHRISPY_DUST2", "S_CHRISPY_BUBBLEDUST",
"S_CHRISPY_CROUCH", "S_CHRISPY_DAB", "S_CHRISPY_HATER", "S_CHRISPY_DROPDASH", "S_CHRISPY_MIDFALL",
"S_CHRISPY_FAN", "S_CHRISPY_DEATH", "S_CHRISPY_DABTRANS", "S_CHRISPY_ACT", "S_CHRISPY_TPOSE",
"S_CHRISPY_NIGHTSFEET", "S_CHRISPY_FLICKY_FLAP", "S_CHRISPY_FLICKY_FALL", "S_CHRISPY_REMERALD_BACK",
"S_CHRISPY_CASH", "S_CHRISPY_COIN")

states[S_CHRISPY_SPARKLE] = {
	sprite = SPR_RSTR,
	frame = FF_ANIMATE|FF_FULLBRIGHT,
	tics = 8,
	var1 = 3,
	var2 = 2
}

states[S_CHRISPY_REVORB] = {
	sprite = SPR_CTHK,
	frame = FF_TRANS30|FF_FULLBRIGHT|FF_ANIMATE,
	tics = 6,
	var1 = 2,
	var2 = 2
}

states[S_CHRISPY_DUST1] = {
	sprite = SPR_NULL,
	tics = 1,
	nextstate = S_CHRISPY_DUST2
}

states[S_CHRISPY_DUST2] = {
	sprite = SPR_CHSP,
	action = A_WaterState,
	var1 = S_CHRISPY_BUBBLEDUST,
	frame = FF_TRANS60
}

states[S_CHRISPY_BUBBLEDUST] = {
	sprite = SPR_BUBL,
	frame = B|FF_TRANS60
}

states[S_CHRISPY_CROUCH] = {
	sprite = SPR_PLAY,
	frame = SPR2_LAND,
	action = A_SetPAnim,
	var2 = PA_IDLE,
	tics = -1
}

states[S_CHRISPY_DAB] = {
	sprite = SPR_PLAY,
	frame = SPR2_POSE,
	tics = -1
}

states[S_CHRISPY_HATER] = {
	sprite = SPR_PLAY,
	frame = SPR2_CNT1,
	tics = 17,
	action = A_Contspeed,
	var2 = PA_DASH, // HAX
	nextstate = S_CHRISPY_HATER
}

states[S_CHRISPY_DROPDASH] = {
	sprite = SPR_PLAY,
	frame = SPR2_TWIN,
	action = A_SetPAnim,
	var2 = PA_ROLL,
	tics = 1,
	nextstate = S_CHRISPY_DROPDASH
}

states[S_CHRISPY_MIDFALL] = {
	sprite = SPR_PLAY,
	frame = SPR2_FLT,
	tics = -1,
	action = A_SetPAnim,
	var2 = PA_FALL,
	nextstate = S_PLAY_FALL
}

states[S_CHRISPY_FAN] = {
	sprite = SPR_PLAY,
	frame = SPR2_FAN_,
	tics = 1,
	var1 = -ANG20,
	nextstate = S_PLAY_SPRING
}

states[S_CHRISPY_DEATH] = {
	sprite = SPR_PLAY,
	frame = SPR2_PDED,
	tics = -1,
	nextstate = S_PLAY_DEAD,
}

states[S_CHRISPY_DABTRANS] = {
	sprite = SPR_PLAY,
	frame = SPR2_TDAB|FF_FULLBRIGHT,
	tics = -1 // this gets modified in CHRISPYCHARS.lua
}

states[S_CHRISPY_ACT] = {
	sprite = SPR_PLAY,
	frame = SPR2_XTRA, // something the player should never be set to
	tics = -1,
	action = A_SetPAnim,
	var2 = PA_IDLE
}

states[S_CHRISPY_TPOSE] = {
	sprite = SPR_PLAY,
	frame = SPR2_TPOS,
	tics = -1,
	action = A_SetPAnim,
	var2 = PA_IDLE
}

states[S_CHRISPY_NIGHTSFEET] = {
	sprite = SPR_NIGH,
	tics = -1,
	var1 = 1
}

states[S_CHRISPY_FLICKY_FLAP] = {
	action = A_ChrispyFlicky,
	sprite = SPR_FL01,
	frame = states[S_FLICKY_01_STAND].frame,
	var1 = states[S_FLICKY_01_STAND].var1,
	var2 = states[S_FLICKY_01_STAND].var2
}

states[S_CHRISPY_FLICKY_FALL] = {
	action = A_ChrispyFlicky,
	sprite = SPR_FL01
}

states[S_CHRISPY_REMERALD_BACK] = {
	sprite = SPR_REMG,
	frame = H | FF_TRANS70 | FF_FULLBRIGHT
}

states[S_CHRISPY_CASH] = {
	sprite = SPR_CHSP,
	frame = D|FF_PAPERSPRITE,
	action = A_RollAngle,
	var1 = 10,
	tics = 1,
	nextstate = S_CHRISPY_CASH,
}

states[S_CHRISPY_COIN] = {
	sprite = SPR_CHSP,
	frame = E,
	tics = 3*TICRATE,
	action = A_RandomFrame,
	var1 = E,
	var2 = H
}

SafeFreeslot("S_POINTY_SPINTRAIL", "S_POINTY_FLIP", "S_POINTY_PEELOUT",
"S_POINTY_BOUNCE1", "S_POINTY_BOUNCE2", "S_POINTY_BOUNCE_LANDING", "S_POINTY_REBOUND",
"S_POINTY_SNEEZE1", "S_POINTY_SNEEZE2", "S_POINTY_TPOSE",
"S_POINTY_DEFAULTDANCE1", "S_POINTY_DEFAULTDANCE2")

states[S_POINTY_SPINTRAIL] = {
	sprite = SPR_CTHK,
	frame = D | FF_TRANS30,
	tics = 6
}

states[S_POINTY_FLIP] = {
	sprite = SPR_PLAY,
	frame = SPR2_FLIP | FF_SPR2ENDSTATE,
	tics = 2,
	var1 = S_CHRISPY_MIDFALL,
	var2 = PA_ABILITY,
	action = A_SetPAnim,
	nextstate = S_POINTY_FLIP
}

states[S_POINTY_BOUNCE1] = {
	sprite = SPR_PLAY,
	frame = SPR2_TWIN|FF_SPR2ENDSTATE,
	tics = 1,
	var1 = S_POINTY_BOUNCE2,
	var2 = PA_ABILITY,
	action = A_SetPAnim,
	nextstate = S_POINTY_BOUNCE1
}

states[S_POINTY_BOUNCE2] = {
	sprite = SPR_PLAY,
	frame = SPR2_BNCE,
	tics = 1,
	var2 = PA_ABILITY,
	action = A_SetPAnim,
	nextstate = S_POINTY_BOUNCE2
}

states[S_POINTY_BOUNCE_LANDING] = {
	sprite = SPR_PLAY,
	frame = SPR2_MLEL|FF_SPR2ENDSTATE,
	tics = 2,
	var1 = S_POINTY_REBOUND,
	var2 = PA_DASH, // hacks hacks hacks hacks hackity hacks
	action = A_SetPAnim,
	nextstate = S_POINTY_BOUNCE_LANDING
}

states[S_POINTY_REBOUND] = {
	sprite = SPR_PLAY,
	frame = SPR2_BNCE,
	tics = 1,
	action = A_SetPAnim,
	var2 = PA_ABILITY,
	nextstate = S_POINTY_FLIP
}

states[S_POINTY_PEELOUT] = {
	sprite = SPR_PLAY,
	frame = SPR2_DASH,
	tics = 1,
	action = A_SetPAnim,
	var2 = PA_DASH,
	nextstate = S_POINTY_PEELOUT
}

states[S_POINTY_SNEEZE1] = {
	sprite = SPR_PLAY,
	frame = SPR2_SNEZ,
	tics = 8,
	var1 = ANG30,
	nextstate = S_POINTY_SNEEZE2
}

states[S_POINTY_SNEEZE2] = {
	sprite = SPR_PLAY,
	frame = SPR2_SNEZ,
	tics = 12,
	var1 = -ANG30,
	nextstate = S_PLAY_HEAD
}

states[S_POINTY_TPOSE] = {
	sprite = SPR_PLAY,
	frame = SPR2_TPOS,
	action = A_SetPAnim,
	var2 = PA_IDLE
}

states[S_POINTY_DEFAULTDANCE1] = {
	sprite = SPR_PLAY,
	frame = SPR2_DFAL|FF_SPR2ENDSTATE,
	tics = 7,
	action = A_SetPAnim,
	var1 = S_POINTY_DEFAULTDANCE2,
	var2 = PA_IDLE,
	nextstate = S_POINTY_DEFAULTDANCE1
}

states[S_POINTY_DEFAULTDANCE2] = {
	sprite = SPR_PLAY,
	frame = SPR2_DFAL,
	tics = TICRATE,
	action = A_SetPAnim,
	var2 = PA_IDLE,
	nextstate = S_PLAY_STND
}

SafeFreeslot("S_FLUFFY_SPINTRAIL", "S_FLUFFY_CARRY", "S_FLUFFY_TOSS1", "S_FLUFFY_TOSS2",
	"S_FLUFFY_GAUGE", "S_FLUFFY_GAUGENEEDLE", "S_FLUFFY_BATTLE")

states[S_FLUFFY_SPINTRAIL] = {
	sprite = SPR_CTHK,
	frame = E | FF_TRANS30,
	tics = 6,
	dispoffset = 1
}

states[S_FLUFFY_CARRY] = {
	sprite = SPR_PLAY,
	frame = SPR2_CARY,
	tics = 1,
	action = A_SetPAnim,
	var2 = PA_ABILITY,
	nextstate = S_PLAY_FLY
}

states[S_FLUFFY_TOSS1] = {
	sprite = SPR_PLAY,
	frame = SPR2_CARY,
	tics = 6,
	action = A_SetPAnim,
	var2 = PA_ABILITY,
	nextstate = S_FLUFFY_TOSS2
}

states[S_FLUFFY_TOSS2] = {
	sprite = SPR_PLAY,
	frame = SPR2_SKID,
	tics = 12,
	action = A_SetPAnim,
	var2 = PA_ABILITY,
	nextstate = S_PLAY_FLY
}

states[S_FLUFFY_GAUGE] = {
	sprite = SPR_CHSP,
	frame = B | FF_FULLBRIGHT,
	var1 = 1
}

states[S_FLUFFY_GAUGENEEDLE] = {
	sprite = SPR_CHSP,
	frame = C | FF_FULLBRIGHT
}

//Battle
states[S_FLUFFY_BATTLE] = {
	sprite = SPR_PLAY,
	frame = SPR2_BATL,
	tics = -1
}

local function MakeFluffyTailState(state, spr2, var1, var2)
	SafeFreeslot(state)
	state = _G[state]
	states[state].sprite = SPR_PLAY
	states[state].frame = spr2
	states[state].nextstate = state
	states[state].tics = -1
	states[state].var1 = var1 or 0 // if != 0, mobj.mirrored is used
	states[state].var2 = var2 or 0 // if != 0, frame number is conserved when switching to and from other states with var2 != 0
end

MakeFluffyTailState("S_FLUFFYTAIL_IDLE", SPR2_TAL0)
MakeFluffyTailState("S_FLUFFYTAIL_TWIRL", SPR2_TAL0, 1)
MakeFluffyTailState("S_FLUFFYTAIL_TWIRL_BELOW", SPR2_TAL2, 0, 1)
MakeFluffyTailState("S_FLUFFYTAIL_TWIRL_ABOVE", SPR2_TAL3, 0, 1)
MakeFluffyTailState("S_FLUFFYTAIL_TWIRL_MIDDLE", SPR2_TAL8, 0, 1)
MakeFluffyTailState("S_FLUFFYTAIL_FLY", SPR2_TAL4)
MakeFluffyTailState("S_FLUFFYTAIL_RUN", SPR2_TAL6)
MakeFluffyTailState("S_FLUFFYTAIL_DAB", SPR2_TAL7)
MakeFluffyTailState("S_FLUFFYTAIL_STOP", SPR2_TAL6, 1)
MakeFluffyTailState("S_FLUFFYTAIL_BLUR", SPR2_TAL5)
MakeFluffyTailState("S_FLUFFYTAIL_TPOSE", SPR2_TAL9)

// Objects

SafeFreeslot(
"MT_CHRISPY_SPARKLE", "MT_CHRISPY_DUST", "MT_CHRISPY_FLICKY", "MT_CHRISPY_COIN",
"MT_POINTY_SPINTRAIL", "MT_POINTY_OVERBALL",
"MT_FLUFFY_SPINTRAIL", "MT_FLUFFY_TAIL")

mobjinfo[MT_CHRISPY_SPARKLE] = {
	spawnstate = S_CHRISPY_SPARKLE,
	height = 8*FRACUNIT,
	radius = 4*FRACUNIT,
	flags = MF_NOCLIP|MF_NOCLIPTHING|MF_NOCLIPHEIGHT|MF_NOBLOCKMAP|MF_NOGRAVITY|MF_SCENERY
}

mobjinfo[MT_CHRISPY_DUST] = {
	spawnstate = S_CHRISPY_DUST1,
	height = mobjinfo[MT_PARTICLE].height,
	radius = mobjinfo[MT_PARTICLE].radius,
	flags = mobjinfo[MT_PARTICLE].flags
}

mobjinfo[MT_CHRISPY_FLICKY] = {
	spawnstate = S_CHRISPY_FLICKY_FLAP,
	seestate = S_CHRISPY_FLICKY_FALL,
	height = mobjinfo[MT_FLICKY_01].height,
	radius = mobjinfo[MT_FLICKY_01].radius,
	speed = 32*FRACUNIT,
	flags = MF_NOGRAVITY|MF_SLIDEME|MF_NOCLIP|MF_NOCLIPHEIGHT
}

mobjinfo[MT_CHRISPY_COIN] = {
	spawnstate = S_CHRISPY_COIN,
	reactiontime = 3,
	seesound = sfx_s227,
	painsound = sfx_s25d,
	height = 4*FRACUNIT,
	radius = 2*FRACUNIT,
	flags = MF_MISSILE|MF_GRENADEBOUNCE|MF_BOUNCE|MF_RUNSPAWNFUNC
}

mobjinfo[MT_POINTY_SPINTRAIL] = {
	spawnstate = S_POINTY_SPINTRAIL,
	height = mobjinfo[MT_THOK].height,
	radius = mobjinfo[MT_THOK].radius,
	flags = mobjinfo[MT_THOK].flags
}

mobjinfo[MT_POINTY_OVERBALL] = {
	spawnstate = S_INVISIBLE,
	height = 2*mobjinfo[MT_PLAYER].height/3,
	radius = mobjinfo[MT_PLAYER].radius,
	flags = MF_NOGRAVITY|MF_NOBLOCKMAP|MF_NOCLIP|MF_NOCLIPTHING|MF_NOCLIPHEIGHT|MF_SCENERY
}

mobjinfo[MT_FLUFFY_SPINTRAIL] = {
	spawnstate = S_FLUFFY_SPINTRAIL,
	height = mobjinfo[MT_THOK].height,
	radius = mobjinfo[MT_THOK].radius,
	flags = mobjinfo[MT_THOK].flags
}

mobjinfo[MT_FLUFFY_TAIL] = {
	spawnstate = S_INVISIBLE,
	height = mobjinfo[MT_PLAYER].height,
	radius = mobjinfo[MT_PLAYER].radius,
	flags = MF_NOGRAVITY|MF_NOBLOCKMAP|MF_NOCLIP|MF_NOCLIPTHING|MF_NOCLIPHEIGHT|MF_SCENERY,
	dispoffset = 2
}